import lms.backend.testing
import lms.model.courses

def test_courses_syllabus_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting a course's syllabus. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
            },
            'A sample course syllabus.',
            None,
        ),

        # Empty
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '120000000'),
            },
            None,
            None,
        ),

        # Miss
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '999'),
            },
            None,
            'Could not resolve course query',
        ),
    ]

    test.base_request_test(test.backend.courses_syllabus_get, test_cases)



import lms.backend.testing
import lms.model.courses
import lms.model.groupsets
import lms.model.testdata.groups

def test_courses_groupsets_memberships_resolve_and_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and listing groupset memberships. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(id = '999'),
            },
            None,
            'Could not resolve group set query',
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '120000000'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(id = '999'),
            },
            None,
            'Could not resolve group set query',
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
            },
            [
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-1'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-2'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-3'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-4'],
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 2'),
            },
            [
                # Note the ordering by group then user.
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-1'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-3'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-2'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-4'],
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 3'),
            },
            [
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groupsets_memberships_resolve_and_list, test_cases)



import typing

BACKEND_TYPE_BLACKBOARD: str = 'blackboard'
BACKEND_TYPE_CANVAS: str = 'canvas'
BACKEND_TYPE_MOODLE: str = 'moodle'

BACKEND_TYPES: typing.List[str] = [
    BACKEND_TYPE_BLACKBOARD,
    BACKEND_TYPE_CANVAS,
    BACKEND_TYPE_MOODLE,
]

HEADER_KEY_BACKEND: str = 'edq-lms-backend'
HEADER_KEY_WRITE: str = 'edq-lms-write'

OUTPUT_FORMAT_JSON: str = 'json'
OUTPUT_FORMAT_TABLE: str = 'table'
OUTPUT_FORMAT_TEXT: str = 'text'

OUTPUT_FORMATS: typing.List[str] = [
    OUTPUT_FORMAT_JSON,
    OUTPUT_FORMAT_TABLE,
    OUTPUT_FORMAT_TEXT,
]



import datetime
import http
import re
import typing

import edq.util.json
import edq.util.net
import edq.util.time
import requests

DEFAULT_PAGE_SIZE: int = 95
HEADER_LINK: str = 'Link'

def fetch_next_canvas_link(response: requests.Response) -> typing.Union[str, None]:
    """
    Fetch the Canvas-style next link within the headers.
    If there is no next link, return None.
    """

    headers = response.headers

    if (HEADER_LINK not in headers):
        return None

    links = headers[HEADER_LINK].split(',')
    for link in links:
        parts = link.split(';')
        if (len(parts) != 2):
            continue

        if (parts[1].strip() != 'rel="next"'):
            continue

        return str(parts[0].strip().strip('<>'))

    return None

def make_request(
        method: str,
        url: str,
        raise_on_404: bool = False,
        json: bool = True,
        **kwargs: typing.Any) -> typing.Union[typing.Any, None]:
    """ Make a single Canvas request and return the decoded JSON body. """

    try:
        _, body_text = edq.util.net.make_request(method, url, **kwargs)
    except requests.HTTPError as ex:
        if (raise_on_404 or (ex.response is None) or (ex.response.status_code != http.HTTPStatus.NOT_FOUND)):
            raise ex

        return None

    if (not json):
        return body_text

    return edq.util.json.loads(body_text, strict = True)

def make_get_request(url: str, **kwargs: typing.Any) -> typing.Union[typing.Any, None]:
    """ Make a single Canvas GET request. """

    return make_request('GET', url, **kwargs)

def make_post_request(url: str, **kwargs: typing.Any) -> typing.Union[typing.Any, None]:
    """ Make a single Canvas POST request. """

    return make_request('POST', url, **kwargs)

def make_delete_request(url: str, **kwargs: typing.Any) -> typing.Union[typing.Any, None]:
    """ Make a single Canvas DELETE request. """

    return make_request('DELETE', url, **kwargs)

def make_get_request_list(
        url: str,
        headers: typing.Dict[str, typing.Any],
        data: typing.Union[typing.Dict[str, typing.Any], None] = None,
        raise_on_404: bool = False,
        ) -> typing.Union[typing.List[typing.Dict[str, typing.Any]], None]:
    """ Repeatedly call make_get_request() (using a JSON body and next link) until there are no more results. """

    output: typing.List[typing.Dict[str, typing.Any]] = []

    next_url: typing.Union[str, None] = url

    while (next_url is not None):
        try:
            response, body_text = edq.util.net.make_get(next_url, headers = headers, data = data)
        except requests.HTTPError as ex:
            if (raise_on_404 or (ex.response is None) or (ex.response.status_code != http.HTTPStatus.NOT_FOUND)):
                raise ex

            return None

        next_url = fetch_next_canvas_link(response)

        new_results = edq.util.json.loads(body_text, strict = True)
        for new_result in new_results:
            output.append(new_result)

    return output

def parse_timestamp(value: typing.Union[str, None]) -> typing.Union[edq.util.time.Timestamp, None]:
    """ Parse a Canvas-style timestamp into a common form. """

    if (value is None):
        return None

    # Parse out some cases that Python <= 3.10 cannot deal with.
    value = re.sub(r'Z$', '+00:00', value)
    value = re.sub(r'(\d\d:\d\d)(\.\d+)', r'\1', value)

    pytime = datetime.datetime.fromisoformat(value)
    return edq.util.time.Timestamp.from_pytime(pytime)



"""
Upload a gradebook (as a table).
"""

import argparse
import ast
import sys

import edq.util.dirent

import lms.backend.instance
import lms.cli.common
import lms.cli.parser
import lms.model.backend
import lms.model.scores

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    gradebook = _load_gradebook(backend, args.path)

    count = backend.courses_gradebook_resolve_and_upload(course_query, gradebook)

    print(f"Uploaded {count} Scores")

    return 0

def _load_gradebook(
        backend: lms.model.backend.APIBackend,
        path: str,
        ) -> lms.model.scores.Gradebook:
    assignments = []
    users = []
    scores = []

    with open(path, 'r', encoding = edq.util.dirent.DEFAULT_ENCODING) as file:
        lineno = 0
        for line in file:
            if (line.strip() == ''):
                continue

            lineno += 1

            parts = [part.strip() for part in line.split("\t")]

            # Process the assignment queries.
            if (lineno == 1):
                if (len(parts) < 2):
                    raise ValueError(f"File '{path}' line {lineno} (assignments line) has the incorrect number of values."
                            + f" Need at least 2 values (user and single assignment), found {len(parts)}.")

                # Skip the users column.
                parts = parts[1:]

                for part in parts:
                    assignment = backend.parse_assignment_query(part)
                    if (assignment is None):
                        raise ValueError(f"File '{path}' line {lineno} has an assignment query that could not be parsed: '{part}'.")

                    assignments.append(assignment)

                continue

            if (len(parts) != (1 + len(assignments))):
                raise ValueError(f"File '{path}' line {lineno} has the incorrect number of values."
                        + f" Expecting {1 + len(assignments)}, found {len(parts)}.")

            # Process user row.
            user = backend.parse_user_query(parts[0])
            if (user is None):
                raise ValueError(f"File '{path}' line {lineno} has an user query that could not be parsed: '{parts[0]}'.")

            users.append(user)

            # User part already processed.
            parts = parts[1:]

            for (i, part) in enumerate(parts):
                part = part.strip()
                if (len(part) == 0):
                    continue

                try:
                    float_score = float(ast.literal_eval(part))
                except Exception:
                    raise ValueError(f"File '{path}' line {lineno} has a score that cannot be converted to a number: '{part}'.")  # pylint: disable=raise-missing-from

                assignment_score = lms.model.scores.AssignmentScore(score = float_score, assignment_query = assignments[i], user_query = user)
                scores.append(assignment_score)

    gradebook = lms.model.scores.Gradebook(assignments, users)

    for score in scores:
        gradebook.add(score)

    return gradebook

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_course = True,
    )

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each row has 2-3 columns: user query, score, and comment (optional).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import copy

import lms.backend.testing
import lms.model.testdata.groups

def test_courses_groups_memberships_add_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of adding group memberships. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '130000000',
                'groupset_id': '131030300',
                'group_id': '131030301',
                'user_ids': [
                    '100060000',
                ],
            },
            1,
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131030300',
                'group_id': '131030301',
                'user_ids': [
                    '100060000',
                    '100070000',
                ],
            },
            2,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groups_memberships_add, test_cases)



"""
The `lms.cli.courses.assignments` package contains tools for interacting with LMS course assignments.
"""

import sys

import edq.util.cli

def main() -> int:
    """ List this CLI dir. """

    return edq.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



import lms.backend.testing
import lms.model.scores
import lms.model.testdata.scores

def test_courses_gradebook_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing a course's gradebook. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'],
            None,
        ),
        (
            {
                'course_id': '120000000',
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Course Using Different Languages'],
            None,
        ),
        (
            {
                'course_id': '130000000',
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Extra Course'],
            None,
        ),

        # Miss
        (
            {
                'course_id': '999',
            },
            None,
            'Could not resolve course query',
        ),
    ]

    test.base_request_test(test.backend.courses_gradebook_list, test_cases)



# pylint: disable=abstract-method

import typing

import lms.backend.canvas.courses.assignments.list
import lms.backend.canvas.courses.assignments.scores.list
import lms.backend.canvas.courses.assignments.scores.upload
import lms.backend.canvas.courses.gradebook.fetch
import lms.backend.canvas.courses.groupsets.create
import lms.backend.canvas.courses.groupsets.delete
import lms.backend.canvas.courses.groupsets.list
import lms.backend.canvas.courses.groupsets.memberships.list
import lms.backend.canvas.courses.groups.create
import lms.backend.canvas.courses.groups.delete
import lms.backend.canvas.courses.groups.list
import lms.backend.canvas.courses.groups.memberships.add
import lms.backend.canvas.courses.groups.memberships.list
import lms.backend.canvas.courses.groups.memberships.subtract
import lms.backend.canvas.courses.list
import lms.backend.canvas.courses.syllabus.fetch
import lms.backend.canvas.courses.users.list
import lms.backend.canvas.courses.users.scores.list
import lms.model.assignments
import lms.model.backend
import lms.model.constants
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.scores
import lms.model.users
import lms.util.parse

class CanvasBackend(lms.model.backend.APIBackend):
    """ An API backend for Instructure's Canvas LMS. """

    def __init__(self,
            server: str,
            auth_token: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(server, lms.model.constants.BACKEND_TYPE_CANVAS, **kwargs)

        if (auth_token is None):
            raise ValueError("Canvas backends require a token.")

        self.auth_token: str = auth_token
        """ The token to authenticate with. """

    def get_standard_headers(self) -> typing.Dict[str, str]:
        headers = super().get_standard_headers()

        headers['Authorization'] = f"Bearer {self.auth_token}"

        return headers

    def courses_list(self,
            **kwargs: typing.Any) -> typing.List[lms.model.courses.Course]:
        return lms.backend.canvas.courses.list.request(self)

    def courses_assignments_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.assignments.Assignment]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.assignments.list.request(self, parsed_course_id)

    def courses_assignments_scores_list(self,
            course_id: str,
            assignment_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_assignment_id = lms.util.parse.required_int(assignment_id, 'assignment_id')
        return lms.backend.canvas.courses.assignments.scores.list.request(self, parsed_course_id, parsed_assignment_id)

    def courses_assignments_scores_upload(self,
            course_id: str,
            assignment_id: str,
            scores: typing.Dict[str, lms.model.scores.ScoreFragment],
            **kwargs: typing.Any) -> int:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_assignment_id = lms.util.parse.required_int(assignment_id, 'assignment_id')
        parsed_scores = {lms.util.parse.required_int(user_id, 'user_id'): score for (user_id, score) in scores.items()}
        return lms.backend.canvas.courses.assignments.scores.upload.request(self, parsed_course_id, parsed_assignment_id, parsed_scores)

    def courses_gradebook_fetch(self,
            course_id: str,
            assignment_ids: typing.Collection[str],
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> lms.model.scores.Gradebook:
        if ((len(assignment_ids) == 0) or (len(user_ids) == 0)):
            assignment_queries = [lms.model.assignments.AssignmentQuery(id = id) for id in assignment_ids]
            user_queries = [lms.model.users.UserQuery(id = id) for id in user_ids]
            return lms.model.scores.Gradebook(assignment_queries, user_queries)

        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_assignment_ids = [lms.util.parse.required_int(assignment_id, 'assignment_id') for assignment_id in assignment_ids]
        parsed_user_ids = [lms.util.parse.required_int(user_id, 'user_id') for user_id in user_ids]

        return lms.backend.canvas.courses.gradebook.fetch.request(self, parsed_course_id, parsed_assignment_ids, parsed_user_ids)

    def courses_groupsets_create(self,
            course_id: str,
            name: str,
            **kwargs: typing.Any) -> lms.model.groupsets.GroupSet:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.groupsets.create.request(self, parsed_course_id, name)

    def courses_groupsets_delete(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> bool:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        return lms.backend.canvas.courses.groupsets.delete.request(self, parsed_course_id, parsed_groupset_id)

    def courses_groupsets_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSet]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.groupsets.list.request(self, parsed_course_id)

    def courses_groupsets_memberships_list(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        return lms.backend.canvas.courses.groupsets.memberships.list.request(self, parsed_course_id, parsed_groupset_id)

    def courses_groups_create(self,
            course_id: str,
            groupset_id: str,
            name: str,
            **kwargs: typing.Any) -> lms.model.groups.Group:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        return lms.backend.canvas.courses.groups.create.request(self, parsed_course_id, parsed_groupset_id, name)

    def courses_groups_delete(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            **kwargs: typing.Any) -> bool:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        parsed_group_id = lms.util.parse.required_int(group_id, 'group_id')
        return lms.backend.canvas.courses.groups.delete.request(self, parsed_course_id, parsed_groupset_id, parsed_group_id)

    def courses_groups_list(self,
            course_id: str,
            groupset_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groups.Group]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        return lms.backend.canvas.courses.groups.list.request(self, parsed_course_id, parsed_groupset_id)

    def courses_groups_memberships_add(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> int:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        parsed_group_id = lms.util.parse.required_int(group_id, 'group_id')
        parsed_user_ids = [lms.util.parse.required_int(user_id, 'user_id') for user_id in user_ids]
        return lms.backend.canvas.courses.groups.memberships.add.request(self, parsed_course_id, parsed_groupset_id, parsed_group_id, parsed_user_ids)

    def courses_groups_memberships_list(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.groupsets.GroupSetMembership]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        parsed_group_id = lms.util.parse.required_int(group_id, 'group_id')
        return lms.backend.canvas.courses.groups.memberships.list.request(self, parsed_course_id, parsed_groupset_id, parsed_group_id)

    def courses_groups_memberships_subtract(self,
            course_id: str,
            groupset_id: str,
            group_id: str,
            user_ids: typing.Collection[str],
            **kwargs: typing.Any) -> int:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_groupset_id = lms.util.parse.required_int(groupset_id, 'groupset_id')
        parsed_group_id = lms.util.parse.required_int(group_id, 'group_id')
        parsed_user_ids = [lms.util.parse.required_int(user_id, 'user_id') for user_id in user_ids]
        return lms.backend.canvas.courses.groups.memberships.subtract.request(self,
                parsed_course_id, parsed_groupset_id, parsed_group_id, parsed_user_ids)

    def courses_syllabus_fetch(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.Union[str, None]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.syllabus.fetch.request(self, parsed_course_id)

    def courses_users_list(self,
            course_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.users.CourseUser]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        return lms.backend.canvas.courses.users.list.request(self, parsed_course_id)

    def courses_users_scores_list(self,
            course_id: str,
            user_id: str,
            **kwargs: typing.Any) -> typing.List[lms.model.scores.AssignmentScore]:
        parsed_course_id = lms.util.parse.required_int(course_id, 'course_id')
        parsed_user_id = lms.util.parse.required_int(user_id, 'user_id')
        return lms.backend.canvas.courses.users.scores.list.request(self, parsed_course_id, parsed_user_id)



import lms.backend.testing
import lms.model.scores
import lms.model.testdata.scores

def test_courses_gradebook_fetch_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of fetching a course's gradebook. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
                'assignment_ids': ['110000100'],
                'user_ids': ['100050000'],
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS_UNRESOLVED['Course 101'],
            None,
        ),

        # Empty
        (
            {
                'course_id': '110000000',
                'assignment_ids': [],
                'user_ids': [],
            },
            lms.model.scores.Gradebook([], []),
            None,
        ),

        # Miss
        (
            {
                'course_id': '999',
                'assignment_ids': [],
                'user_ids': [],
            },
            lms.model.scores.Gradebook([], []),
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_gradebook_fetch, test_cases)



"""
Upload scores (and optional comments) for an assignment.
"""

import argparse
import ast
import sys
import typing

import edq.util.dirent

import lms.backend.instance
import lms.cli.common
import lms.cli.parser
import lms.model.backend
import lms.model.scores
import lms.model.users

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    assignment_query = lms.cli.common.check_required_assignment(backend, config)
    if (assignment_query is None):
        return 2

    scores = _load_scores(backend, args.path, args.skip_rows)

    count = backend.courses_assignments_scores_resolve_and_upload(course_query, assignment_query, scores)

    print(f"Uploaded {count} Scores")

    return 0

def _load_scores(
        backend: lms.model.backend.APIBackend,
        path: str,
        skip_rows: bool,
        ) -> typing.Dict[lms.model.users.UserQuery, lms.model.scores.ScoreFragment]:
    scores = {}

    with open(path, 'r', encoding = edq.util.dirent.DEFAULT_ENCODING) as file:
        lineno = 0
        real_rows = 0
        for line in file:
            lineno += 1

            if (line.strip() == ''):
                continue

            real_rows += 1

            if (real_rows <= skip_rows):
                continue

            parts = [part.strip() for part in line.split("\t")]
            if (len(parts) not in [2, 3]):
                raise ValueError(f"File '{path}' line {lineno} has the incorrect number of values. Expecting 2-3, found {len(parts)}.")

            user_query = backend.parse_user_query(parts[0])
            if (user_query is None):
                raise ValueError(f"File '{path}' line {lineno} has a user query that could not be parsed: '{parts[0]}'.")

            score = None
            if (parts[1] != ''):
                try:
                    score = float(ast.literal_eval(parts[1]))
                except Exception:
                    raise ValueError(f"File '{path}' line {lineno} has a score that cannot be converted to a number: '{parts[1]}'.")  # pylint: disable=raise-missing-from

            comment = None
            if (len(parts) == 3):
                comment = parts[2]

            scores[user_query] = lms.model.scores.ScoreFragment(score = score, comment = comment)

    return scores

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_course = True,
            include_assignment = True,
            include_skip_rows = True,
    )

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each row has 2-3 columns: user query, score, and comment (optional).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import typing

import lms.backend.canvas.common

BASE_ENDPOINT = "/api/v1/courses/{course_id}"

def request(backend: typing.Any,
        course_id: int,
        ) -> typing.Union[str, None]:
    """ Fetch a course syllabus. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id)
    headers = backend.get_standard_headers()

    data = {
        'include': ['syllabus_body'],
    }

    raw_object = lms.backend.canvas.common.make_get_request(url, headers = headers, data = data)
    if (raw_object is None):
        identifiers = {
            'course_id': course_id,
        }
        backend.not_found('fetch syllabus', identifiers)

        return None

    result = raw_object.get('syllabus_body', None)
    if (result is None):
        return None

    return str(result)



import copy

import lms.backend.testing
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.users
import lms.model.testdata.groups

def test_courses_groups_memberships_resolve_and_subtract_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and subtracting group memberships. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'group_query': lms.model.groups.GroupQuery(name = 'Group 1-1'),
                'user_queries': [
                    lms.model.users.UserQuery(name = 'extra-course-student-1'),
                ],
            },
            (1, False),
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'group_query': lms.model.groups.GroupQuery(name = 'Group 1-1'),
                'user_queries': [
                    lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    lms.model.users.UserQuery(name = 'extra-course-student-2'),
                ],
            },
            (2, False),
            None,
        ),

        # Delete Empty
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'group_query': lms.model.groups.GroupQuery(name = 'Group 1-1'),
                'delete_empty': True,
                'user_queries': [
                    lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    lms.model.users.UserQuery(name = 'extra-course-student-2'),
                ],
            },
            (2, True),
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groups_memberships_resolve_and_subtract, test_cases)



import typing

import edq.util.net
import requests

import lms.backend.blackboard.backend
import lms.backend.canvas.backend
import lms.backend.moodle.backend
import lms.model.constants
import lms.model.backend

def get_backend(
        server: typing.Union[str, None] = None,
        backend_type: typing.Union[str, None] = None,
        **kwargs: typing.Any) -> lms.model.backend.APIBackend:
    """
    Get an instance of an API backend from the given information.
    If the backend type is not explicitly provided,
    this function will attempt to guess it from other information.
    """

    if (server is None):
        raise ValueError("No LMS server address provided.")

    server = server.strip()
    if (not server.startswith('http')):
        server = 'http://' + server

    backend_type = guess_backend_type(server, backend_type = backend_type)
    if (backend_type is None):
        raise ValueError(f"Unable to guess backend type from server: '{server}'.")

    if (backend_type == lms.model.constants.BACKEND_TYPE_CANVAS):
        return lms.backend.canvas.backend.CanvasBackend(server, **kwargs)

    if (backend_type == lms.model.constants.BACKEND_TYPE_MOODLE):
        return lms.backend.moodle.backend.MoodleBackend(server, **kwargs)

    if (backend_type == lms.model.constants.BACKEND_TYPE_BLACKBOARD):
        return lms.backend.blackboard.backend.BlackboardBackend(server, **kwargs)

    if (backend_type in lms.model.constants.BACKEND_TYPES):
        raise ValueError(f"Instance creation not yet supported for backend type: '{backend_type}'.")

    raise ValueError(f"Unknown backend type: '{backend_type}'. Known backend types: {lms.model.constants.BACKEND_TYPES}.")

def guess_backend_type(
        server: typing.Union[str, None] = None,
        backend_type: typing.Union[str, None] = None,
        **kwargs: typing.Any) -> typing.Union[str, None]:
    """
    Attempt to guess the backend type from a server.
    This function will return None it cannot guess the backend type.
    """

    if (backend_type is not None):
        return backend_type

    if (server is None):
        return None

    # Try looking at the URL string itself.
    backend_type = guess_backend_type_from_url(server)
    if (backend_type is not None):
        return backend_type

    # Make a request to the server and examine the response.
    backend_type = guess_backend_type_from_request(server)
    if (backend_type is not None):
        return backend_type

    return None

def guess_backend_type_from_request(server: str, timeout_secs: float = edq.util.net.DEFAULT_REQUEST_TIMEOUT_SECS) -> typing.Union[str, None]:
    """
    Attempt to guess the backend type by pinging the server.
    This function will not do any lexical analysis on the server string.
    """

    options = {
        'allow_redirects': False,
    }

    try:
        response, _ = edq.util.net.make_get(server,
                raise_for_status = False,
                timeout_secs = timeout_secs,
                additional_requests_options = options)
    except requests.exceptions.ConnectionError:
        return None
    except requests.exceptions.Timeout:
        return None

    header_keys = [key.lower() for key in response.headers.keys()]

    # Blackboard sends a special header.
    if ('x-blackboard-product' in header_keys):
        return lms.model.constants.BACKEND_TYPE_BLACKBOARD

    # Canvas sends a special header.
    if ('x-canvas-meta' in header_keys):
        return lms.model.constants.BACKEND_TYPE_CANVAS

    # Canvas requests that a specific cookie is set.
    if ('_normandy_session' in response.headers.get('set-cookie', '')):
        return lms.model.constants.BACKEND_TYPE_CANVAS

    # Moodle will try to redirect with a special header.
    if (response.headers.get('x-redirect-by', '').lower() == 'moodle'):
        return lms.model.constants.BACKEND_TYPE_MOODLE

    # Moodle requests that a specific cookie is set.
    if ('MoodleSession' in response.headers.get('set-cookie', '')):
        return lms.model.constants.BACKEND_TYPE_MOODLE

    return None

def guess_backend_type_from_url(server: str) -> typing.Union[str, None]:
    """
    Attempt to guess the backend type only from a string server URL.
    This function will only do lexical analysis on the string (no HTTP requests will be made).
    """

    server = server.lower().strip()

    if ('canvas' in server):
        return lms.model.constants.BACKEND_TYPE_CANVAS

    if ('moodle' in server):
        return lms.model.constants.BACKEND_TYPE_MOODLE

    if ('blackboard' in server):
        return lms.model.constants.BACKEND_TYPE_BLACKBOARD

    return None



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.courses

BASE_ENDPOINT = "/api/v1/courses?per_page={page_size}"

def request(backend: typing.Any,
        ) -> typing.List[lms.model.courses.Course]:
    """ List courses. """

    url = backend.server + BASE_ENDPOINT.format(page_size = lms.backend.canvas.common.DEFAULT_PAGE_SIZE)
    headers = backend.get_standard_headers()

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers)
    if (raw_objects is None):
        backend.not_found('list courses', {})

        return []

    return [lms.backend.canvas.model.course(raw_object) for raw_object in raw_objects]



import lms.backend.testing

def test_courses_groupsets_delete_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of deleting group sets. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '130000000',
                'groupset_id': '131010100',
            },
            True,
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '999',
            },
            False,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groupsets_delete, test_cases)



"""
List the current configuration options.
"""

import argparse
import sys

import edq.cli.config.list

import lms.cli.parser

CONFIG_FIELD_SEPARATOR: str = "\t"

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    return int(edq.cli.config.list.run_cli(args))

def main() -> int:
    """ Get a parser, parse the args, and call run. """

    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get a parser and add addition flags. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_server = False,
            include_auth = False,
    )

    edq.cli.config.list.modify_parser(parser)

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.constants
import lms.model.groupsets

BASE_ENDPOINT = "/api/v1/courses/{course_id}/group_categories"

def request(backend: typing.Any,
        course_id: int,
        name: str,
        ) -> lms.model.groupsets.GroupSet:
    """ Create a group set. """

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id)
    headers = backend.get_standard_headers()

    headers[lms.model.constants.HEADER_KEY_WRITE] = 'true'

    data = {
        'name': name,
    }

    raw_object = lms.backend.canvas.common.make_post_request(url, headers = headers, data = data)
    if (raw_object is None):
        identifiers = {
            'course_id': course_id,
        }
        backend.not_found('create group set', identifiers)

        raise ValueError(f"Unable to create group set '{name}' for course '{course_id}'.")

    return lms.backend.canvas.model.group_set(raw_object)



import typing

import lms.backend.canvas.common
import lms.backend.canvas.model
import lms.model.groups

BASE_ENDPOINT = "/api/v1/group_categories/{groupset_id}/groups?per_page={page_size}"

def request(backend: typing.Any,
        course_id: int,
        groupset_id: int,
        ) -> typing.List[lms.model.groups.Group]:
    """ List a group set's groups. """

    url = backend.server + BASE_ENDPOINT.format(groupset_id = groupset_id, page_size = lms.backend.canvas.common.DEFAULT_PAGE_SIZE)
    headers = backend.get_standard_headers()

    raw_objects = lms.backend.canvas.common.make_get_request_list(url, headers = headers)
    if (raw_objects is None):
        identifiers = {
            'course_id': course_id,
            'groupset_id': groupset_id,
        }
        backend.not_found('list groups', identifiers)

        return []

    return [lms.backend.canvas.model.group(raw_object) for raw_object in raw_objects]



import lms.backend.testing
import lms.model.assignments
import lms.model.courses
import lms.model.scores
import lms.model.testdata.assignments
import lms.model.testdata.users
import lms.model.testdata.scores
import lms.model.users

def test_courses_gradebook_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting a course's gradebook. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'assignment_queries': lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'].assignments,
                'user_queries': lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'].users,
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'],
            None,
        ),

        # Course Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
                'assignment_queries': lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'].assignments,
                'user_queries': lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'].users,
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'],
            None,
        ),

        # No Queries
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'assignment_queries': [],
                'user_queries': [],
            },
            lms.model.testdata.scores.COURSE_GRADEBOOKS['Course 101'],
            None,
        ),

        # Miss
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '999'),
                'assignment_queries': [],
                'user_queries': [],
            },
            None,
            'Could not resolve course query',
        ),

        # Single Assignment
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'assignment_queries': [
                    lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
                ],
                'user_queries': [],
            },
            _build_gradebook('Extra Course', ['Assignment 2'], None),
            None,
        ),

        # Single User
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'assignment_queries': [],
                'user_queries': [
                    lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
                ],
            },
            _build_gradebook('Extra Course', None, ['extra-course-student-2']),
            None,
        ),

        # Doubles
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
                'assignment_queries': [
                    lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
                    lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 3'].to_query(),
                ],
                'user_queries': [
                    lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
                    lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'].to_query(),
                ],
            },
            _build_gradebook('Extra Course', ['Assignment 1', 'Assignment 3'], ['extra-course-student-1', 'extra-course-student-4']),
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_gradebook_get, test_cases)

def _build_gradebook(course_name, allowed_assignment_names, allowed_user_names):
    assignment_queries = []
    for assignment in sorted(lms.model.testdata.assignments.COURSE_ASSIGNMENTS[course_name].values()):
        if ((allowed_assignment_names is None) or (assignment.name in allowed_assignment_names)):
            assignment_queries.append(assignment.to_query())

    user_queries = []
    for user in sorted(lms.model.testdata.users.COURSE_USERS[course_name].values()):
        if (not user.is_student()):
            continue

        if ((allowed_user_names is None) or (user.name in allowed_user_names)):
            user_queries.append(user.to_query())

    gradebook = lms.model.scores.Gradebook(assignment_queries, user_queries)

    for (assignment_name, user_scores) in lms.model.testdata.scores.COURSE_ASSIGNMENT_SCORES[course_name].items():
        if ((allowed_assignment_names is not None) and (assignment_name not in allowed_assignment_names)):
            continue

        for (user_name, score) in user_scores.items():
            if ((allowed_user_names is not None) and (user_name not in allowed_user_names)):
                continue

            gradebook.add(score)

    return gradebook



import typing

import edq.util.json
import edq.util.time

import lms.model.assignments
import lms.model.base
import lms.model.users

class ScoreFragment(edq.util.json.DictConverter):
    """ A small subset of information about a score. """

    def __init__(self,
            score: typing.Union[float, None] = None,
            comment: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        self.score: typing.Union[float, None] = score
        """
        The numeric score.
        A None value indicates that the score should be cleared.
        """

        self.comment: typing.Union[str, None] = comment
        """ An optional comment for this score. """

class AssignmentScore(lms.model.base.BaseType):
    """
    The score assignment to a student for an assignment (or scorable object).
    """

    CORE_FIELDS = [
        'id', 'user_query', 'assignment_query', 'score', 'submission_date', 'graded_date', 'comment',
    ]

    def __init__(self,
            id: typing.Union[str, None] = None,
            score: typing.Union[float, None] = None,
            submission_date: typing.Union[edq.util.time.Timestamp, None] = None,
            graded_date: typing.Union[edq.util.time.Timestamp, None] = None,
            comment: typing.Union[str, None] = None,
            assignment_query: typing.Union[lms.model.assignments.AssignmentQuery, None] = None,
            user_query: typing.Union[lms.model.users.UserQuery, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.id: typing.Union[str, None] = id
        """ The LMS's identifier for this score. """

        self.assignment_query: typing.Union[lms.model.assignments.AssignmentQuery, None] = assignment_query
        """ The assignment associated with this score. """

        self.user_query: typing.Union[lms.model.users.UserQuery, None] = user_query
        """ The user associated with this score. """

        self.score: typing.Union[float, None] = score
        """ The assignment score. """

        self.submission_date: typing.Union[edq.util.time.Timestamp, None] = submission_date
        """ The datetime that the submission that received this score was submitted. """

        self.graded_date: typing.Union[edq.util.time.Timestamp, None] = graded_date
        """ The datetime that the submission that received this score was graded. """

        self.comment: typing.Union[str, None] = comment
        """ A comment attached to this score. """

    def to_fragment(self) -> ScoreFragment:
        """ Get a score fragment from this assignment score. """

        return ScoreFragment(score = self.score, comment = self.comment)

class Gradebook(lms.model.base.BaseType):
    """
    A gradebook that contains scores for a set of users and assignments.
    """

    CORE_FIELDS = [
        'assignments', 'users', '_entries',
    ]

    def __init__(self,
            assignments: typing.List[lms.model.assignments.AssignmentQuery],
            users: typing.List[lms.model.users.UserQuery],
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.assignments: typing.List[lms.model.assignments.AssignmentQuery] = assignments
        """ The assignments represented in this gradebook. """

        self.users: typing.List[lms.model.users.UserQuery] = users
        """ The users represented in this gradebook. """

        self._entries: typing.Dict[str, AssignmentScore] = {}
        """ The scores held by this gradebook. """

    def _make_key(self, assignment_query: lms.model.assignments.AssignmentQuery, user_query: lms.model.users.UserQuery) -> str:
        """ Create a key for this gradebook entry. """

        return f"{assignment_query.id}::{user_query.id}"

    def get(self,
            assignment_query: lms.model.assignments.AssignmentQuery,
            user_query: lms.model.users.UserQuery,
            ) -> typing.Union[AssignmentScore, None]:
        """ Get the target gradebook entry. """

        found_assignment = None
        for assignment in self.assignments:
            if (assignment.match(assignment_query)):
                found_assignment = assignment
                break

        if (found_assignment is None):
            return None

        found_user = None
        for user in self.users:
            if (user.match(user_query)):
                found_user = user
                break

        if (found_user is None):
            return None

        return self._entries.get(self._make_key(found_assignment, found_user), None)

    def get_scores_by_assignment(self,
            ) -> typing.Dict[
                lms.model.assignments.AssignmentQuery,
                typing.Dict[lms.model.users.UserQuery, AssignmentScore]]:
        """ Get all entries indexed by assignment. """

        results: typing.Dict[
                lms.model.assignments.AssignmentQuery,
                typing.Dict[lms.model.users.UserQuery, AssignmentScore]] = {}

        for assignment in self.assignments:
            results[assignment] = {}

            for user in self.users:
                key = self._make_key(assignment, user)
                if (key not in self._entries):
                    continue

                results[assignment][user] = self._entries[key]

        return results

    def add(self, score: AssignmentScore) -> None:
        """
        Add the score to this gradebook.
        If the user or assignment is not already in this gradebook, raise an exception.

        The gradebook takes ownership of the score.
        """

        found_assignment = None
        for assignment in self.assignments:
            if (assignment.match(score.assignment_query)):
                found_assignment = assignment
                break

        if (found_assignment is None):
            raise ValueError(f"Could not match gradebook assignment to score's assignment '{score.assignment_query}'.")

        found_user = None
        for user in self.users:
            if (user.match(score.user_query)):
                found_user = user
                break

        if (found_user is None):
            raise ValueError(f"Could not match gradebook user to score's user '{score.user_query}'.")

        # Update the score's queries.
        score.assignment_query = found_assignment
        score.user_query = found_user

        self._entries[self._make_key(found_assignment, found_user)] = score

    def update_queries(self,
            assignment_queries: typing.List[lms.model.assignments.ResolvedAssignmentQuery],
            user_queries: typing.List[lms.model.users.ResolvedUserQuery],
            ) -> None:
        """ Update any assignment/user queries with the supplied ones (matching on ID). """

        assignment_map = {query.id: query for query in assignment_queries}
        user_map = {query.id: query for query in user_queries}

        self.assignments = [assignment_map.get(assignment.id, assignment) for assignment in self.assignments]
        self.users = [user_map.get(user.id, user) for user in self.users]

        for score in self._entries.values():
            if (score.assignment_query is not None):
                score.assignment_query = assignment_map.get(score.assignment_query.id, score.assignment_query)

            if (score.user_query is not None):
                score.user_query = user_map.get(score.user_query.id, score.user_query)

    def as_text_rows(self,
            skip_headers: bool = False,
            pretty_headers: bool = False,
            separator: str = ': ',
            empty_value: str = '',
            **kwargs: typing.Any) -> typing.List[str]:
        rows: typing.List[str] = []
        for user in sorted(self.users):
            # Add in a user separator.
            if (len(rows) > 0):
                rows.append('')

            rows.append(f"User{separator}{user}")

            for assignment in sorted(self.assignments):
                score = self.get(assignment, user)

                text = empty_value
                if (score is not None):
                    text = str(score.score)

                rows.append(f"{assignment}{separator}{text}")

        return rows

    def get_headers(self,
            pretty_headers: bool = False,
            **kwargs: typing.Any) -> typing.List[str]:
        return ['User'] + [str(query) for query in sorted(self.assignments)]

    def as_table_rows(self,
            empty_value: str = '',
            **kwargs: typing.Any) -> typing.List[typing.List[str]]:
        rows = []
        for user in sorted(self.users):
            row = [str(user)]
            for assignment in sorted(self.assignments):
                score = self.get(assignment, user)

                text = empty_value
                if (score is not None):
                    text = str(score.score)

                row.append(text)

            rows.append(row)

        return rows

    def as_json_dict(self,
        include_extra_fields: bool = False,
            **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        """
        Get a dict representation of this object meant for display as JSON.
        (Note that we are not returning JSON, just a dict that is ready to be converted to JSON.)
        Calling this method differs from passing this object to json.dumps() (or any sibling),
        because this method may not include all fields, may flatten or alter fields, and will order fields differently.
        """

        scores = []
        for assignment in sorted(self.assignments):
            row = []
            for user in sorted(self.users):
                score: typing.Any = self.get(assignment, user)
                if (score is not None):
                    score = score.as_json_dict(include_extra_fields = include_extra_fields, **kwargs)

                row.append(score)

            scores.append(row)

        return {
            'assignments': self.assignments,
            'users': self.users,
            'scores_assignment_user': scores,
        }

    def __len__(self) -> int:
        return len(self._entries)



import typing

import lms.backend.canvas.common
import lms.model.constants
import lms.model.scores

BASE_ENDPOINT = "/api/v1/courses/{course_id}/assignments/{assignment_id}/submissions/update_grades"

def request(backend: typing.Any,
        course_id: int,
        assignment_id: int,
        scores: typing.Dict[int, lms.model.scores.ScoreFragment],
        ) -> int:
    """ Upload scores and return the number of scores sent. """

    if (len(scores) == 0):
        return 0

    url = backend.server + BASE_ENDPOINT.format(course_id = course_id, assignment_id = assignment_id)
    headers = backend.get_standard_headers()

    headers[lms.model.constants.HEADER_KEY_WRITE] = 'true'

    data = {}
    for (user_id, score) in scores.items():
        text_score = ''
        if (score.score is not None):
            text_score = str(score.score)

        data[f"grade_data[{user_id}][posted_grade]"] = text_score

        if ((score.comment is not None) and (len(score.comment) > 0)):
            data[f"grade_data[{user_id}][text_comment]"] = score.comment

    response = lms.backend.canvas.common.make_post_request(url, headers = headers, data = data)
    if (response is None):
        return 0

    return len(scores)



import typing

import edq.util.time

import lms.model.base
import lms.model.query

class AssignmentQuery(lms.model.query.BaseQuery):
    """
    A class for the different ways one can attempt to reference an LMS assignment.
    In general, an assignment can be queried by:
     - LMS Assignment ID (`id`)
     - Full Name (`name`)
     - f"{name} ({id})"
    """

    _include_email = False

class ResolvedAssignmentQuery(lms.model.query.ResolvedBaseQuery, AssignmentQuery):
    """
    A AssignmentQuery that has been resolved (verified) from a real assignment instance.
    """

    _include_email = False

    def __init__(self,
            assignment: 'Assignment',
            **kwargs: typing.Any) -> None:
        super().__init__(id = assignment.id, name = assignment.name, **kwargs)

class Assignment(lms.model.base.BaseType):
    """
    An assignment within a course.
    """

    CORE_FIELDS = [
        'id', 'name', 'description',
        'open_date', 'close_date', 'due_date',
        'points_possible',
    ]

    def __init__(self,
            id: typing.Union[str, int, None] = None,
            name: typing.Union[str, None] = None,
            description: typing.Union[str, None] = None,
            open_date: typing.Union[edq.util.time.Timestamp, None] = None,
            close_date: typing.Union[edq.util.time.Timestamp, None] = None,
            due_date: typing.Union[edq.util.time.Timestamp, None] = None,
            points_possible: typing.Union[float, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        if (id is None):
            raise ValueError("Assignment must have an id.")

        self.id: str = str(id)
        """ The LMS's identifier for this assignment. """

        self.name: typing.Union[str, None] = name
        """ The display name of this assignment. """

        self.description: typing.Union[str, None] = description
        """ The description of this assignment. """

        self.open_date: typing.Union[edq.util.time.Timestamp, None] = open_date
        """ The datetime that this assignment becomes open at. """

        self.close_date: typing.Union[edq.util.time.Timestamp, None] = close_date
        """ The datetime that this assignment becomes close at. """

        self.due_date: typing.Union[edq.util.time.Timestamp, None] = due_date
        """ The datetime that this assignment is due at. """

        self.points_possible: typing.Union[float, None] = points_possible
        """ The maximum number of points possible for this assignment. """

    def to_query(self) -> ResolvedAssignmentQuery:
        """ Get a query representation of this assignment. """

        return ResolvedAssignmentQuery(self)



import lms.backend.testing
import lms.model.courses
import lms.model.scores
import lms.model.testdata.assignments
import lms.model.testdata.scores
import lms.model.testdata.users

def test_courses_gradebook_upload_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of uploading gradebooks. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '130000000',
                'gradebook': _get_base_gradebook(),
            },
            4,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_gradebook_upload, test_cases)

def test_courses_gradebook_resolve_and_upload_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and uploading gradebooks. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'gradebook': _get_base_gradebook(),
            },
            4,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_gradebook_resolve_and_upload, test_cases)

def _get_base_gradebook():
    base_assignments = [
        lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 1'].to_query(),
        lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Extra Course']['Assignment 2'].to_query(),
    ]

    base_users = [
        lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'].to_query(),
        lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'].to_query(),
    ]

    base_gradebook = lms.model.scores.Gradebook(base_assignments, base_users)

    base_gradebook.add(lms.model.scores.AssignmentScore(score = 1.0,
            assignment_query = base_assignments[0], user_query = base_users[0]))
    base_gradebook.add(lms.model.scores.AssignmentScore(score = 0.5,
            assignment_query = base_assignments[0], user_query = base_users[1]))
    base_gradebook.add(lms.model.scores.AssignmentScore(score = 1.0, comment = 'extra-course-student-1 comment',
            assignment_query = base_assignments[1], user_query = base_users[0]))
    base_gradebook.add(lms.model.scores.AssignmentScore(score = 0.5, comment = 'extra-course-student-2 comment',
            assignment_query = base_assignments[1], user_query = base_users[1]))

    return base_gradebook



import typing

import lms.model.courses

# {course_name: course, ...}
COURSES: typing.Dict[str, lms.model.courses.Course] = {
    'Course 101': lms.model.courses.Course(
        id = 110000000,
        name = 'Course 101',
    ),
    'Course Using Different Languages': lms.model.courses.Course(
        id = 120000000,
        name = 'Course Using Different Languages',
    ),
    'Extra Course': lms.model.courses.Course(
        id = 130000000,
        name = 'Extra Course',
    ),
}



import lms.backend.testing
import lms.model.testdata.scores

def test_courses_users_scores_fetch_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of fetching user's scores. """

    scores = lms.model.testdata.scores.COURSE_ASSIGNMENT_SCORES_UNRESOLVED

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
                'user_id': '100050000',
            },
            scores['Course 101']['Homework 0']['course-student'],
            None,
        ),

        # Miss
        (
            {
                'course_id': '110000000',
                'assignment_id': '999',
                'user_id': '100050000',
            },
            None,
            None,
        ),
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
                'user_id': '999',
            },
            None,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_scores_fetch, test_cases)



import lms.backend.testing
import lms.model.courses
import lms.model.testdata.courses

def test_courses_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting courses. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Empty
        (
            {
                'course_queries': [],
            },
            [
            ],
            None,
        ),

        # Base - List
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(id = '120000000'),
                    lms.model.courses.CourseQuery(id = '130000000'),
                ],
            },
            [
                lms.model.testdata.courses.COURSES['Course Using Different Languages'],
                lms.model.testdata.courses.COURSES['Extra Course'],
            ],
            None,
        ),

        # Base - Fetch
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(id = '120000000'),
                ],
            },
            [
                lms.model.testdata.courses.COURSES['Course Using Different Languages'],
            ],
            None,
        ),

        # Query - Name
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(name = 'Course 101'),
                ],
            },
            [
                lms.model.testdata.courses.COURSES['Course 101'],
            ],
            None,
        ),

        # Query - Label
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(name = 'Course 101', id = '110000000'),
                ],
            },
            [
                lms.model.testdata.courses.COURSES['Course 101'],
            ],
            None,
        ),

        # Miss - ID
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(id = 999),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Query
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Miss - Partial Match
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(id = '120000000', name = 'ZZZ'),
                ],
            },
            [
            ],
            None,
        ),

        # Multiple Match
        (
            {
                'course_queries': [
                    lms.model.courses.CourseQuery(id = '110000000'),
                    lms.model.courses.CourseQuery(name = 'Course 101'),
                ],
            },
            [
                lms.model.testdata.courses.COURSES['Course 101'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_get, test_cases)



import lms.backend.testing
import lms.model.courses
import lms.model.testdata.groupsets

def test_courses_groupsets_resolve_and_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and listing course groupsets. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
            },
            [
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '120000000'),
            },
            [
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'],
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 3'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
            },
            [
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 1'],
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 2'],
                lms.model.testdata.groupsets.COURSE_GROUPSETS['Extra Course']['Group Set 3'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_groupsets_resolve_and_list, test_cases)



"""
Get a course's syllabus.
"""

import argparse
import sys

import lms.backend.instance
import lms.cli.common
import lms.cli.parser

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    syllabus = backend.courses_syllabus_get(course_query)

    if (syllabus is None):
        print(f"<No syllabus found for course: '{course_query}'.>")
    else:
        print(syllabus)

    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_course = True,
    )

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import lms.backend.testing
import lms.model.testdata.assignments

def test_courses_assignments_fetch_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of fetching course assignments. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
            },
            lms.model.testdata.assignments.COURSE_ASSIGNMENTS['Course 101']['Homework 0'],
            None,
        ),

        # Miss
        (
            {
                'course_id': '110000000',
                'assignment_id': '999',
            },
            None,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_assignments_fetch, test_cases)



import lms.backend.testing
import lms.model.assignments
import lms.model.courses
import lms.model.testdata.scores
import lms.model.users

def test_courses_users_scores_get_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of getting users' scores. """

    scores = lms.model.testdata.scores.COURSE_ASSIGNMENT_SCORES

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Empty
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
                'assignment_queries': [],
            },
            [
            ],
            None,
        ),

        # Base
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
                'assignment_queries': [
                    lms.model.assignments.AssignmentQuery(id = '110000100'),
                ],
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),

        # Course Query
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
                'user_query': lms.model.users.UserQuery(id = '100050000'),
                'assignment_queries': [
                    lms.model.assignments.AssignmentQuery(id = '110000100'),
                ],
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),

        # Queries
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
                'user_query': lms.model.users.UserQuery(email = 'course-student@test.edulinq.org'),
                'assignment_queries': [
                    lms.model.assignments.AssignmentQuery(name = 'Homework 0'),
                ],
            },
            [
                scores['Course 101']['Homework 0']['course-student'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_scores_get, test_cases)



import re
import typing

import lms.model.courses
import lms.model.users

COURSE_ROLE_MAPPING: typing.Dict[str, lms.model.users.CourseRole] = {
    'Guest': lms.model.users.CourseRole.OTHER,
    'Student': lms.model.users.CourseRole.STUDENT,
    'Grader': lms.model.users.CourseRole.GRADER,
    'TeachingAssistant': lms.model.users.CourseRole.ADMIN,
    'Instructor': lms.model.users.CourseRole.OWNER,
}

def parse_id(text: str) -> str:
    """ Blackboard tends to put other text around their ids. """

    text = text.strip()

    match = re.match(r'^_(\d+)_\d*', text)
    if (match is not None):
        return match.group(1)

    return text

def format_id(id: int) -> str:
    """ Format a Blackboard primary ID. """

    return f"_{id}_1"

def course(data: typing.Dict[str, typing.Any]) -> lms.model.courses.Course:
    """
    Create a Blackboard course from raw data.
    """

    return lms.model.courses.Course(
        id = parse_id(data['id']),
        name = data['name'],
    )

def course_user(data: typing.Dict[str, typing.Any]) -> lms.model.users.CourseUser:
    """
    Create a Blackboard course from raw data.
    """

    user = data['user']

    return lms.model.users.CourseUser(
        id = parse_id(user['id']),
        name = user['name']['given'],
        email = user['contact']['email'],
        raw_role = data['courseRoleId'],
        role = COURSE_ROLE_MAPPING.get(data['courseRoleId'], None),
    )



import typing

import lms.backend.canvas.common
import lms.model.assignments
import lms.model.backend
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.scores
import lms.model.users
import lms.util.parse

ENROLLMENT_TYPE_TO_ROLE: typing.Dict[str, lms.model.users.CourseRole] = {
    'ObserverEnrollment': lms.model.users.CourseRole.OTHER,
    'StudentEnrollment': lms.model.users.CourseRole.STUDENT,
    'TaEnrollment': lms.model.users.CourseRole.GRADER,
    'DesignerEnrollment': lms.model.users.CourseRole.ADMIN,
    'TeacherEnrollment': lms.model.users.CourseRole.OWNER,
}
"""
Canvas enrollment types mapped to roles.
This map is ordered by priority/power.
The later in the dict, the more power.
"""

_testing_override: bool = False  # pylint: disable=invalid-name
""" A special override to signal testing. """

def assignment(data: typing.Dict[str, typing.Any]) -> lms.model.assignments.Assignment:
    """
    Create a Canvas assignment associated with a course.

    See: https://developerdocs.instructure.com/services/canvas/resources/assignments
    """

    for field in ['id']:
        if (field not in data):
            raise ValueError(f"Canvas assignment is missing '{field}' field.")

    # Modify specific arguments before creation.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')
    data['due_date'] = lms.backend.canvas.common.parse_timestamp(data.get('due_at', None))
    data['open_date'] = lms.backend.canvas.common.parse_timestamp(data.get('unlock_at', None))
    data['close_date'] = lms.backend.canvas.common.parse_timestamp(data.get('lock_at', None))

    return lms.model.assignments.Assignment(**data)

def assignment_score(data: typing.Dict[str, typing.Any]) -> lms.model.scores.AssignmentScore:
    """
    Create a Canvas assignment score.

    See: https://developerdocs.instructure.com/services/canvas/resources/scores
    """

    # Check for important fields.
    for field in ['id', 'assignment_id', 'user_id']:
        if (field not in data):
            raise ValueError(f"Canvas assignment score is missing '{field}' field.")

    # Modify specific arguments before creation.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')
    data['score'] = lms.util.parse.optional_float(data.get('score', None), 'score')
    data['points_possible'] = lms.util.parse.optional_float(data.get('points_possible', None), 'points_possible')
    data['submission_date'] = lms.backend.canvas.common.parse_timestamp(data.get('submitted_at', None))
    data['graded_date'] = lms.backend.canvas.common.parse_timestamp(data.get('graded_at', None))

    assignment_id = lms.util.parse.required_string(data.get('assignment_id', None), 'assignment_id')
    data['assignment_query'] = lms.model.assignments.AssignmentQuery(id = assignment_id)

    user_id = lms.util.parse.required_string(data.get('user_id', None), 'user_id')
    data['user_query'] = lms.model.users.UserQuery(id = user_id)

    return lms.model.scores.AssignmentScore(**data)

def course(data: typing.Dict[str, typing.Any]) -> lms.model.courses.Course:
    """
    Create a Canvas course.

    See: https://developerdocs.instructure.com/services/canvas/resources/courses
    """

    # Check for important fields.
    for field in ['id']:
        if (field not in data):
            raise ValueError(f"Canvas course is missing '{field}' field.")

    # Modify specific arguments before creation.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')

    return lms.model.courses.Course(**data)

def course_user(backend: lms.model.backend.APIBackend, data: typing.Dict[str, typing.Any]) -> lms.model.users.CourseUser:
    """
    Create a Canvas user associated with a course.

    See: https://developerdocs.instructure.com/services/canvas/resources/users
    """

    # Check for important fields.
    for field in ['id']:
        if (field not in data):
            raise ValueError(f"Canvas user is missing '{field}' field.")

    # Modify specific arguments before sending them to super.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')

    # Canvas sometimes has email under different fields.
    if ((data.get('email', None) is None) or (len(data.get('email', '')) == 0)):
        data['email'] = data.get('login_id', None)

    enrollments = data.get('enrollments', None)
    if (enrollments is not None):
        data['raw_role'] = _parse_role_from_enrollments(enrollments)
        data['role'] = ENROLLMENT_TYPE_TO_ROLE.get(data['raw_role'], None)

        # Canvas has a discontinuity with its default course roles.
        # We need to patch this during testing.
        if ((backend.is_testing() or _testing_override) and data['email'] == 'course-admin@test.edulinq.org'):
            data['role'] = lms.model.users.CourseRole.ADMIN

    return lms.model.users.CourseUser(**data)

def group(data: typing.Dict[str, typing.Any]) -> lms.model.groups.Group:
    """
    Create a Canvas group associated with a course.

    See: https://developerdocs.instructure.com/services/canvas/resources/groups
    """

    # Check for important fields.
    for field in ['id']:
        if (field not in data):
            raise ValueError(f"Canvas group is missing '{field}' field.")

    # Modify specific arguments before creation.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')

    return lms.model.groups.Group(**data)

def group_set(data: typing.Dict[str, typing.Any]) -> lms.model.groupsets.GroupSet:
    """
    Create a Canvas group set associated with a course.

    See: https://developerdocs.instructure.com/services/canvas/resources/group_categories
    """

    # Check for important fields.
    for field in ['id']:
        if (field not in data):
            raise ValueError(f"Canvas group set is missing '{field}' field.")

    # Modify specific arguments before creation.
    data['id'] = lms.util.parse.required_string(data.get('id', None), 'id')

    return lms.model.groupsets.GroupSet(**data)

def _parse_role_from_enrollments(enrollments: typing.Any) -> typing.Union[str, None]:
    """
    Try to parse the user's role from their enrollments.
    If multiple roles are discovered, take the "highest" one.

    See: https://developerdocs.instructure.com/services/canvas/resources/enrollments
    """

    if (not isinstance(enrollments, list)):
        return None

    best_role = None
    best_index = -1

    enrollment_types = list(ENROLLMENT_TYPE_TO_ROLE.keys())

    for enrollment in enrollments:
        if (not isinstance(enrollment, dict)):
            continue

        if (enrollment.get('enrollment_state', None) != 'active'):
            continue

        role = enrollment.get('role', None)

        role_index = -1
        if (role in enrollment_types):
            role_index = enrollment_types.index(role)

        if ((best_role is None) or (role_index > best_index)):
            best_role = role
            best_index = role_index

    return best_role



import typing

import lms.model.base
import lms.model.query

class CourseQuery(lms.model.query.BaseQuery):
    """
    A class for the different ways one can attempt to reference an LMS course.
    In general, a course can be queried by:
     - LMS Course ID (`id`)
     - Full Name (`name`)
     - f"{name} ({id})"
    """

    _include_email = False

class ResolvedCourseQuery(lms.model.query.ResolvedBaseQuery, CourseQuery):
    """
    A CourseQuery that has been resolved (verified) from a real course instance.
    """

    _include_email = False

    def __init__(self,
            course: 'Course',
            **kwargs: typing.Any) -> None:
        super().__init__(id = course.id, name = course.name, **kwargs)

class Course(lms.model.base.BaseType):
    """
    A course.
    """

    CORE_FIELDS = [
        'id', 'name',
    ]

    def __init__(self,
            id: typing.Union[str, int, None] = None,
            name: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        if (id is None):
            raise ValueError("Course must have an id.")

        self.id: str = str(id)
        """ The LMS's identifier for this course. """

        self.name: typing.Union[str, None] = name
        """ The display name of this course. """

    def to_query(self) -> ResolvedCourseQuery:
        """ Get a query representation of this course. """

        return ResolvedCourseQuery(self)



import lms.backend.testing
import lms.model.groups
import lms.model.groupsets

def test_courses_groups_resolve_and_delete_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of deleting groups (with resolution). """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'group_query': lms.model.groups.GroupQuery(name = 'Group 1-1'),
            },
            True,
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'group_query': lms.model.groups.GroupQuery(name = 'ZZZ'),
            },
            None,
            'Could not resolve group query',
        ),
    ]

    test.base_request_test(test.backend.courses_groups_resolve_and_delete, test_cases)



import logging
import typing

import edq.testing.serverrunner
import edq.util.net
import edq.util.parse
import edq.util.reflection

import lms.backend.instance
import lms.cli.parser
import lms.model.constants
import lms.util.net

BACKEND_REQUEST_CLEANING_FUNCS: typing.Dict[str, typing.Callable] = {
    lms.model.constants.BACKEND_TYPE_BLACKBOARD: lms.util.net.clean_blackboard_response,
    lms.model.constants.BACKEND_TYPE_CANVAS: lms.util.net.clean_canvas_response,
    lms.model.constants.BACKEND_TYPE_MOODLE: lms.util.net.clean_moodle_response,
}

class LMSServerRunner(edq.testing.serverrunner.ServerRunner):
    """ A server runner specifically for LMS servers. """

    def __init__(self,
            backend_type: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.backend_type: typing.Union[str, None] = backend_type
        """
        The type of server being run.
        This value will be resolved after the server is started
        (since part of resolution may involve pinging the server.
        """

        self._old_exchanges_clean_func: typing.Union[str, None] = None
        """
        The value of edq.util.net._exchanges_clean_func when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._old_set_exchanges_clean_func: bool = False
        """
        The value of lms.cli.parser._set_exchanges_clean_func when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._old_make_request_exchange_complete_func: typing.Union[edq.util.net.HTTPExchangeComplete, None] = None
        """
        The value of edq.util.net._make_request_exchange_complete_func when start() is called.
        The original value may be changed in start(), and will be reset in stop().
        """

        self._old_serverrunner_logging_level: typing.Union[int, None] = None
        """
        The logging level for the edq server runner befoe tests are run.
        The original value may be changed in start(), and will be reset in stop().
        """

    def start(self) -> None:
        super().start()

        # Resolve the backend type.
        self.backend_type = lms.backend.instance.guess_backend_type(self.server, backend_type = self.backend_type)
        if (self.backend_type is None):
            raise ValueError(f"Unable to determine backend type for server '{self.server}'.")

        # Set configs.

        exchange_clean_func = BACKEND_REQUEST_CLEANING_FUNCS.get(self.backend_type, lms.util.net.clean_lms_response)
        exchange_clean_func_name = edq.util.reflection.get_qualified_name(exchange_clean_func)

        self._old_exchanges_clean_func = edq.util.net._exchanges_clean_func
        edq.util.net._exchanges_clean_func = exchange_clean_func_name

        self._old_set_exchanges_clean_func = lms.cli.parser._set_exchanges_clean_func
        lms.cli.parser._set_exchanges_clean_func = False

        def _make_request_callback(exchange: edq.util.net.HTTPExchange) -> None:
            # Restart if the request is a write.
            if (edq.util.parse.boolean(exchange.headers.get(lms.model.constants.HEADER_KEY_WRITE, False))):
                self.restart()

        self._old_make_request_exchange_complete_func = edq.util.net._make_request_exchange_complete_func
        edq.util.net._make_request_exchange_complete_func = typing.cast(edq.util.net.HTTPExchangeComplete, _make_request_callback)

        # Disable logging from the runner, since it may disrupt CLI tests.
        logger = logging.getLogger('edq.testing.serverrunner')
        self._old_serverrunner_logging_level = logger.level
        logger.setLevel(logging.WARNING)

    def stop(self) -> bool:
        if (self._old_serverrunner_logging_level is not None):
            logger = logging.getLogger('edq.testing.serverrunner')
            logger.setLevel(self._old_serverrunner_logging_level)
            self._old_serverrunner_logging_level = None

        if (not super().stop()):
            return False

        # Restore old configs.

        edq.util.net._exchanges_clean_func = self._old_exchanges_clean_func
        self._old_exchanges_clean_func = None

        lms.cli.parser._set_exchanges_clean_func = self._old_set_exchanges_clean_func
        self._old_set_exchanges_clean_func = False

        edq.util.net._make_request_exchange_complete_func = self._old_make_request_exchange_complete_func
        self._old_make_request_exchange_complete_func = None

        return True

    def identify_server(self) ->  bool:
        backend_type = lms.backend.instance.guess_backend_type_from_request(self.server, timeout_secs = self.identify_wait_secs)
        return (backend_type is not None)



import lms.backend.testing

def test_courses_syllabus_fetch_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of fetching a course's syllabus. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
            },
            'A sample course syllabus.',
            None,
        ),

        # Empty
        (
            {
                'course_id': '120000000',
            },
            None,
            None,
        ),

        # Miss
        (
            {
                'course_id': '999',
            },
            None,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_syllabus_fetch, test_cases)



"""
Customize an argument parser for LMS Toolkit.
"""

import argparse
import typing

import edq.core.argparser
import edq.util.net
import edq.util.reflection

import lms
import lms.model.constants
import lms.util.net

CONFIG_FILENAME: str = 'edq-lms.json'

DEFAULT_SKIP_ROWS: int = 0

_set_exchanges_clean_func: bool = True  # pylint: disable=invalid-name
"""
Whether to set the exchanges clean function when creating the parser.
This may be disabled for testing.
"""

def get_parser(description: str,
        include_server: bool = True,
        include_auth: bool = True,
        include_output_format: bool = False,
        include_course: bool = False,
        include_assignment: bool = False,
        include_user: bool = False,
        include_groupset: bool = False,
        include_group: bool = False,
        include_net: bool = True,
        include_skip_rows: bool = False,
        ) -> argparse.ArgumentParser:
    """
    Get an argument parser specialized for LMS Toolkit.
    """

    config_options = {
        'config_filename': CONFIG_FILENAME,
        'cli_arg_config_map': {
            'server': 'server',
            'backend_type': 'backend_type',
            'auth_user': 'auth_user',
            'auth_password': 'auth_password',
            'auth_token': 'auth_token',
            'course': 'course',
            'assignment': 'assignment',
            'user': 'user',
            'group': 'group',
            'groupset': 'groupset',
        },
    }

    parser = edq.core.argparser.get_default_parser(
            description,
            version = f"v{lms.__version__}",
            include_net = include_net,
            config_options = config_options,
    )

    # Ensure that responses are cleaned as LMS responses.
    if (include_net):
        if (_set_exchanges_clean_func):
            edq.util.net._exchanges_clean_func = edq.util.reflection.get_qualified_name(lms.util.net.clean_lms_response)

    if (include_server):
        parser.add_argument('--server', dest = 'server',
            action = 'store', type = str, default = None,
            help = 'The address of the LMS server to connect to.')

        parser.add_argument('--server-type', dest = 'backend_type',
            action = 'store', type = str,
            default = None, choices = lms.model.constants.BACKEND_TYPES,
            help = 'The type of LMS being connected to (this can normally be guessed from the server address).')

    if (include_auth):
        parser.add_argument('--auth-user', dest = 'auth_user',
            action = 'store', type = str, default = None,
            help = 'The user to authenticate with.')

        parser.add_argument('--auth-password', dest = 'auth_password',
            action = 'store', type = str, default = None,
            help = 'The password to authenticate with.')

        parser.add_argument('--auth-token', dest = 'auth_token',
            action = 'store', type = str, default = None,
            help = 'The token to authenticate with.')

    if (include_course):
        parser.add_argument('--course', dest = 'course',
            action = 'store', type = str, default = None,
            help = 'The course to target for this operation.')

    if (include_assignment):
        parser.add_argument('--assignment', dest = 'assignment',
            action = 'store', type = str, default = None,
            help = 'The assignment to target for this operation.')

    if (include_user):
        parser.add_argument('--user', dest = 'user',
            action = 'store', type = str, default = None,
            help = 'The user to target for this operation.')

    if (include_groupset):
        parser.add_argument('--groupset', dest = 'groupset',
            action = 'store', type = str, default = None,
            help = 'The group set to target for this operation.')

    if (include_group):
        parser.add_argument('--group', dest = 'group',
            action = 'store', type = str, default = None,
            help = 'The group to target for this operation.')

    if (include_output_format):
        parser.add_argument('--format', dest = 'output_format',
            action = 'store', type = str,
            default = lms.model.constants.OUTPUT_FORMAT_TEXT, choices = lms.model.constants.OUTPUT_FORMATS,
            help = 'The format to display the output as (default: %(default)s).')

        parser.add_argument('--skip-headers', dest = 'skip_headers',
            action = 'store_true', default = False,
            help = 'Skip headers when outputting results, will not apply to all formats (default: %(default)s).')

        parser.add_argument('--pretty-headers', dest = 'pretty_headers',
            action = 'store_true', default = False,
            help = 'When displaying headers, try to make them look "pretty" (default: %(default)s).')

        parser.add_argument('--include-extra-fields', dest = 'include_extra_fields',
            action = 'store_true', default = False,
            help = 'Include non-common (usually LMS-specific) fields in results (default: %(default)s).')

    if (include_skip_rows):
        parser.add_argument('--skip-rows', dest = 'skip_rows',
            action = 'store', type = int, default = DEFAULT_SKIP_ROWS,
            help = 'The number of header rows to skip (default: %(default)s).')

    return typing.cast(argparse.ArgumentParser, parser)



import typing

import lms.backend.canvas.common
import lms.model.groups
import lms.model.groupsets
import lms.model.users

BASE_ENDPOINT = "/api/v1/group_categories/{groupset_id}/export"

def request(backend: typing.Any,
        course_id: int,
        groupset_id: int,
        ) -> typing.List[lms.model.groupsets.GroupSetMembership]:
    """ List group set memberships. """

    url = backend.server + BASE_ENDPOINT.format(groupset_id = groupset_id)
    headers = backend.get_standard_headers()

    raw_text = lms.backend.canvas.common.make_get_request(url, headers = headers, json = False)
    if (raw_text is None):
        identifiers = {
            'course_id': course_id,
            'groupset_id': groupset_id,
        }
        backend.not_found('list group set memberships', identifiers)

        return []

    raw_objects = _parse_csv(raw_text)

    groupset_query = lms.model.groupsets.GroupSetQuery(id = groupset_id)

    memberships = []
    for raw_object in raw_objects:
        # Canvas includes unassigned users.
        if (raw_object.get('canvas_group_id', None) is None):
            continue

        user_query = lms.model.users.UserQuery(
            id = raw_object.get('canvas_user_id', None),
            name = raw_object.get('name', None),
            email = raw_object.get('user_id', None)
        )
        group_query = lms.model.groups.GroupQuery(
            id = raw_object.get('canvas_group_id', None),
            name = raw_object.get('group_name', None),
        )

        membership = lms.model.groupsets.GroupSetMembership(user = user_query, groupset = groupset_query, group = group_query)
        memberships.append(membership)

    return memberships

def _parse_csv(text: str) -> typing.List[typing.Dict[str, str]]:
    rows: typing.List[typing.Dict[str, str]] = []
    headers = None

    for line in text.splitlines():
        if (len(line.strip()) == 0):
            continue

        parts = [part.strip() for part in line.split(',')]

        if (headers is None):
            headers = parts
            continue

        if (len(parts) != len(headers)):
            raise ValueError("Canvas returned an improperly formatted CSV file.")

        rows.append({headers[i]: parts[i] for i in range(len(headers))})

    return rows



import edq.testing.unittest

import lms.model.query

class TestQuery(edq.testing.unittest.BaseTest):
    """ Test queries. """

    def test_parse_int_query_base(self):
        """ Test parsing an int-id query. """

        # [(text, check email, expected), ...]
        test_cases = [
            # None
            (
                None,
                True,
                None,
            ),

            # Empty
            (
                '',
                True,
                None,
            ),

            # Whitespace
            (
                '  ',
                True,
                None,
            ),

            # ID
            (
                '123',
                True,
                lms.model.query.BaseQuery(id = '123'),
            ),

            # Name
            (
                'name',
                True,
                lms.model.query.BaseQuery(name = 'name'),
            ),

            # Email
            (
                'email@test.edulinq.org',
                True,
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
            ),

            # Label - Name
            (
                'name (123)',
                True,
                lms.model.query.BaseQuery(id = '123', name = 'name'),
            ),

            # Label - Email
            (
                'email@test.edulinq.org (123)',
                True,
                lms.model.query.BaseQuery(id = '123', email = 'email@test.edulinq.org'),
            ),

            # Don't Check Email - ID
            (
                '123',
                False,
                lms.model.query.BaseQuery(id = '123'),
            ),

            # Don't Check Email - Name
            (
                'name',
                False,
                lms.model.query.BaseQuery(name = 'name'),
            ),

            # Don't Check Email - Email
            (
                'email@test.edulinq.org',
                False,
                lms.model.query.BaseQuery(name = 'email@test.edulinq.org'),
            ),

            # Don't Check Email - Label - Name
            (
                'name (123)',
                False,
                lms.model.query.BaseQuery(id = '123', name = 'name'),
            ),

            # Don't Check Email - Label - Email
            (
                'email@test.edulinq.org (123)',
                False,
                lms.model.query.BaseQuery(id = '123', name = 'email@test.edulinq.org'),
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (text, check_email, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{text}'):"):
                actual = lms.model.query.parse_int_query(lms.model.query.BaseQuery, text, check_email)
                self.assertJSONEqual(expected, actual)

    def test_match_base(self):
        """ Test base matching. """

        # [(query, target, expected), ...]
        test_cases = [
            # Exact Match
            (
                lms.model.query.BaseQuery(id = '123'),
                lms.model.query.BaseQuery(id = '123'),
                True,
            ),
            (
                lms.model.query.BaseQuery(name = 'name'),
                lms.model.query.BaseQuery(name = 'name'),
                True,
            ),
            (
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
                True,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name'),
                lms.model.query.BaseQuery(id = '123', name = 'name'),
                True,
            ),
            (
                lms.model.query.BaseQuery(id = '123', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', email = 'email@test.edulinq.org'),
                True,
            ),
            (
                lms.model.query.BaseQuery(name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(name = 'name', email = 'email@test.edulinq.org'),
                True,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                True,
            ),

            # Subset Match
            (
                lms.model.query.BaseQuery(id = '123'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                True,
            ),
            (
                lms.model.query.BaseQuery(name = 'name'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                True,
            ),
            (
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                True,
            ),

            # Incomplete Mismatch
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(name = 'name'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
                False,
            ),

            # Partial Mismatch
            (
                lms.model.query.BaseQuery(id = '999', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'ZZZZ', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'ZZZZZ@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                False,
            ),

            # Full Mismatch
            (
                lms.model.query.BaseQuery(id = '123'),
                lms.model.query.BaseQuery(id = '999'),
                False,
            ),
            (
                lms.model.query.BaseQuery(name = 'name'),
                lms.model.query.BaseQuery(name = 'ZZZZ'),
                False,
            ),
            (
                lms.model.query.BaseQuery(email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(email = 'ZZZZZ@test.edulinq.org'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name'),
                lms.model.query.BaseQuery(id = '999', name = 'ZZZZ'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '999', email = 'ZZZZZ@test.edulinq.org'),
                False,
            ),
            (
                lms.model.query.BaseQuery(name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(name = 'ZZZZ', email = 'ZZZZZ@test.edulinq.org'),
                False,
            ),
            (
                lms.model.query.BaseQuery(id = '123', name = 'name', email = 'email@test.edulinq.org'),
                lms.model.query.BaseQuery(id = '999', name = 'ZZZZ', email = 'ZZZZZ@test.edulinq.org'),
                False,
            ),
        ]

        for (i, test_case) in enumerate(test_cases):
            (query, target, expected) = test_case

            with self.subTest(msg = f"Case {i} ('{query}' vs '{target}'):"):
                actual = query.match(target)
                self.assertEqual(expected, actual)



import typing

import lms.model.base
import lms.model.query
import lms.model.users

class GroupQuery(lms.model.query.BaseQuery):
    """
    A class for the different ways one can attempt to reference an LMS group.
    In general, a group can be queried by:
     - LMS Group ID (`id`)
     - Full Name (`name`)
     - f"{name} ({id})"
    """

    _include_email = False

class ResolvedGroupQuery(lms.model.query.ResolvedBaseQuery, GroupQuery):
    """
    A GroupQuery that has been resolved (verified) from a real group instance.
    """

    _include_email = False

    def __init__(self,
            group: 'Group',
            **kwargs: typing.Any) -> None:
        super().__init__(id = group.id, name = group.name, **kwargs)

class Group(lms.model.base.BaseType):
    """
    A formal collection of users.
    """

    CORE_FIELDS = [
        'id', 'name',
    ]

    def __init__(self,
            id: typing.Union[str, int, None] = None,
            name: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        if (id is None):
            raise ValueError("Groups must have an id.")

        self.id: str = str(id)
        """ The LMS's identifier for this group. """

        self.name: typing.Union[str, None] = name
        """ The display name of this group. """

    def to_query(self) -> ResolvedGroupQuery:
        """ Get a query representation of this group. """

        return ResolvedGroupQuery(self)

class GroupMembership(lms.model.base.BaseType):
    """
    An instance of a user being in a group.
    """

    CORE_FIELDS = [
        'group', 'user',
    ]

    def __init__(self,
            user: lms.model.users.UserQuery,
            group: GroupQuery,
            **kwargs: typing.Any) -> None:
        super().__init__(**kwargs)

        self.group: GroupQuery = group
        """ The group the user belongs to. """

        self.user: lms.model.users.UserQuery = user
        """ The user in a group. """

    def update_queries(self,
            users: typing.Union[typing.Dict[str, lms.model.users.ResolvedUserQuery], None] = None,
            groups: typing.Union[typing.Dict[str, ResolvedGroupQuery], None] = None,
            ) -> None:
        """
        Update the queries with resolved variants.
        The maps should be keyed by respective ids.
        """

        if ((users is not None) and (self.user.id in users)):
            self.user = users[self.user.id]

        if ((groups is not None) and (self.group.id in groups)):
            self.group = groups[self.group.id]



import re
import typing

INT_PATTERN: re.Pattern = re.compile(r'^\d+$')

def compare_maybe_ints(a: typing.Union[str, int, None], b: typing.Union[str, int, None]) -> int:
    """
    Perform a standard comparison (-1, 0, 1) on two strings that are presumed to be integers.
    If they are not integers, compare them as strings.
    """

    if ((a is None) or (b is None)):
        if ((a is None) and (b is None)):
            return 0

        if (a is None):
            return -1

        return 1

    values = [a, b]
    has_str = False

    for (i, value) in enumerate(values):
        if (isinstance(value, int)):
            continue

        value = str(value)
        if (INT_PATTERN.match(value) is not None):
            value = int(value)
        else:
            has_str = True

        values[i] = value

    if (has_str):
        values = [str(value) for value in values]

    if (values[0] < values[1]):  # type: ignore[operator]
        return -1

    if (values[0] > values[1]):  # type: ignore[operator]
        return 1

    return 0



import lms.backend.testing
import lms.model.courses
import lms.model.testdata.users

def test_courses_users_resolve_and_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and listing course users. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '110000000'),
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-grader'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-other'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-owner'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-student'],
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '120000000'),
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-grader'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-other'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-owner'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-student'],
            ],
            None,
        ),
        (
            {
                'course_query': lms.model.courses.CourseQuery(id = '130000000'),
            },
            [
                lms.model.testdata.users.COURSE_USERS['Extra Course']['course-owner'],
                lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-1'],
                lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-2'],
                lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-3'],
                lms.model.testdata.users.COURSE_USERS['Extra Course']['extra-course-student-4'],
            ],
            None,
        ),

        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
            },
            [
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-admin'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-grader'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-other'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-owner'],
                lms.model.testdata.users.COURSE_USERS['Course 101']['course-student'],
            ],
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_users_resolve_and_list, test_cases)



import copy

import lms.backend.testing
import lms.model.groupsets

DUMMY_ID: str = '123456789'

def test_courses_groupsets_resolve_and_create_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of creating group sets (with resolution). """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Course 101'),
                'name': 'test_groupset_1',
            },
            lms.model.groupsets.GroupSet(id = DUMMY_ID, name = 'test_groupset_1'),
            None,
        ),
    ]

    # IDs from the backend may be inconsistent.
    def _clean_result(result):
        result = copy.deepcopy(result)
        result.id = DUMMY_ID

        return result

    test.base_request_test(test.backend.courses_groupsets_resolve_and_create, test_cases,
            actual_clean_func = _clean_result)



# pylint: disable=abstract-method

import logging
import typing
import urllib.parse

import bs4
import edq.util.net

import lms.model.backend
import lms.model.constants
import lms.util.net

class MoodleBackend(lms.model.backend.APIBackend):
    """ An API backend for the Moodle LMS. """

    def __init__(self,
            server: str,
            auth_user: typing.Union[str, None] = None,
            auth_password: typing.Union[str, None] = None,
            **kwargs: typing.Any) -> None:
        super().__init__(server, lms.model.constants.BACKEND_TYPE_MOODLE, **kwargs)

        if (auth_user is None):
            raise ValueError("Moodle backends require a username.")

        if (auth_password is None):
            raise ValueError("Moodle backends require a password.")

        self._username = auth_user
        """ The username to authenticate with. """

        self._password = auth_password
        """ The password to authenticate with. """

        self._session_headers: typing.Union[typing.Dict[str, typing.Any], None] = None
        """ The headers (e.g., cookies) for our logged in Moodle session. """

    def _login(self, update_server: bool = True) -> None:
        """
        Try to login to the Moodle server.
        If `update_server` is true, then this may try to update the backend's server location if redirected by the Moodle server.
        """

        # Check if we are already logged in.
        if (self._session_headers is not None):
            return

        response, body = edq.util.net.make_get(self.server + '/login/index.php')
        cookies = lms.util.net.parse_cookies(response.headers.get('set-cookie', None))

        new_cookies = {
            'MoodleSession': cookies['moodlesession'],
        }
        text_cookies = '; '.join(['='.join(items) for items in new_cookies.items()])

        # Parse the login token from the page HTML.
        document = bs4.BeautifulSoup(body, 'html.parser')
        token = document.select('input[name="logintoken"]')[0]['value']

        headers = {
            'cookie': text_cookies,
            'host': urllib.parse.urlparse(self.server).netloc,
        }
        data = {
            'anchor': '',
            'logintoken': token,
            'username': self._username,
            'password': self._password,
        }

        response, _ = edq.util.net.make_post(self.server + '/login/index.php',
                headers = headers, data = data,
                allow_redirects = False)

        # Check for a successful login.
        cookies = lms.util.net.parse_cookies(response.headers.get('set-cookie', None))
        if ('moodleid1_' in cookies):
            self._session_headers = {
                'cookie': response.headers.get('set-cookie', None),
                # Insert a header to identify the user.
                'edq-lms-moodle-user': self._username,
            }
            return

        # Login Failed

        # The specified server/host needs to match exactly what the Moodle server wants it to be,
        # e.g., `127.0.0.1` does not work when the sevrer wants the host to be `localhost`.
        # If these do not match, we will get a redirect here.
        # Use this redirect to discover the correct server.
        location = response.headers.get('location', None)
        if (update_server and (location is not None) and (not location.startswith(self.server))):
            parts = urllib.parse.urlparse(location)
            host = f"{parts.scheme}://{parts.netloc}"

            logging.debug(("Mismatch in the client-specified server ('%s') and server-requested host ('%s')."
                    + " To avoid extra requests, update the server (e.g., `--server`) to match the host."),
                    self.server, host)

            # Update the server and try to login again (without updating the server again (to avoid loops)).
            self.server = host
            self._login(update_server = False)
            return

        raise ValueError(f"Could not log into Moodle server ({self.server}) with user '{self._username}'. Is username/password correct?")



# pylint: disable=invalid-name

"""
Generate test data by starting the specified server and running all tests in this project.
"""

import argparse
import sys

import edq.testing.serverrunner

import lms.cli.parser
import lms.testing.testdata

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    return lms.testing.testdata.generate(args)

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    parser = lms.cli.parser.get_parser(__doc__.strip(),
            include_server = False,
            include_auth = False,
    )

    edq.testing.serverrunner.modify_parser(parser)

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import os
import typing

import edq.testing.cli

import lms.backend.testing
import lms.model.constants

THIS_DIR: str = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
ROOT_DIR: str = os.path.join(THIS_DIR, '..', '..', '..')

CANVAS_TEST_EXCHANGES_DIR: str = os.path.join(ROOT_DIR, 'testdata', 'lms-docker-canvas-testdata', 'testdata', 'http')

USER_TOKENS: typing.Dict[str, str] = {
    'course-admin@test.edulinq.org': 'nHRkQ39czXL2x7xxKrPYmvtYTyWJCCHCVRMZTfTfZtJZZWXHnkN9UhnCy37XuYeK',
    'course-grader@test.edulinq.org': 'uwRzuhDzDyEJBuJ8QR8PRTLAZHRU7ErY6aTtACNtB7tHZNVzLLw2AGZTGLQya9YX',
    'course-other@test.edulinq.org': 'VntnXWUfHDYDGhFV8VPmUrMEVuwJ3JeJ898FFDf7DHkGJ7vmrEW3eJx9cuHukh94',
    'course-owner@test.edulinq.org': 'xkC8V8BWX4RFx7JMYyZuyDvtDAKRxuGHRxTR268eHzXCPYU46vw89DrBADat4n6U',
    'course-student@test.edulinq.org': 'T7J3DQMJzamkcPVWtkRh6zczx7CHEBy3JGJkvEeavcQyVKDGL9MAkveyJyDuAUEL',
    'server-admin@test.edulinq.org': 'hHKUya4aDV6BnPDuDv8rL7TBFmxmGuBzTMFRrmFfDNaZM4Wy7WQKfufNt9kW9m3W',
    'server-creator@test.edulinq.org': 'R9Z9GhYLrUArnc2cTAeu3Q7fkBhw7CZtuKB8A9eTVhvHWFKWrDVD769GnNzraAGJ',
    'server-owner@test.edulinq.org': 'ycY4AhQ8ZGwk7L38vGur9HtG2WXevMcRh62eXU8KAfGRuXaXhXZE2wCthWVzRZn2',
    'server-user@test.edulinq.org': 'vYNF6mWfcz4mQYBG6XJXeJh8x4WNNeQHkEkVDWAQxc8JBC9GJFwCffP9fznK4QMK',
}
"""
The tokens for each known user.
These tokens were taken from the Canvas test data repo:
https://github.com/edulinq/lms-docker-canvas-testdata
"""

DEFAULT_TOKEN: str = USER_TOKENS['course-owner@test.edulinq.org']
""" The default token for these tests. """

class CanvasBackendTest(lms.backend.testing.BackendTest):
    """ A backend test for Canvas. """

    @classmethod
    def child_class_setup(cls) -> None:
        cls.server_key = lms.model.constants.BACKEND_TYPE_CANVAS

        cls.backend_type = lms.model.constants.BACKEND_TYPE_CANVAS

        cls.exchanges_dir = CANVAS_TEST_EXCHANGES_DIR

        cls.backend_args.update({
            'auth_token': DEFAULT_TOKEN,
        })

        cls.params_to_skip += [
            'per_page',
        ]

        cls.headers_to_skip += [
        ]

    def modify_cli_test_info(self, test_info: edq.testing.cli.CLITestInfo) -> None:
        super().modify_cli_test_info(test_info)

        test_info.arguments += [
            '--auth-token', self.backend.auth_token,
        ]

    def set_user(self, email: str) -> None:
        super().set_user(email)

        # Update the token the backend is using.
        self.backend.auth_token = USER_TOKENS[email]

    def clear_user(self) -> None:
        super().clear_user()

        # Switch back to the default token.
        self.backend.auth_token = DEFAULT_TOKEN

# Attatch tests to this class.
lms.backend.testing.attach_test_cases(CanvasBackendTest)



import copy

import lms.backend.testing
import lms.model.testdata.groups

def test_courses_groups_memberships_list_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of listing group memberships. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        (
            {
                'course_id': '110000000',
                'groupset_id': '999',
                'group_id': '999',
            },
            [
            ],
            None,
        ),
        (
            {
                'course_id': '120000000',
                'groupset_id': '999',
                'group_id': '999',
            },
            [
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131010100',
                'group_id': '131010101',
            },
            [
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-1'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-2'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131010100',
                'group_id': '131010102',
            },
            [
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-3'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 1']['extra-course-student-4'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131020200',
                'group_id': '131020201',
            },
            [
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-1'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-3'],
            ],
            None,
        ),
        (
            {
                'course_id': '130000000',
                'groupset_id': '131020200',
                'group_id': '131020202',
            },
            [
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-2'],
                lms.model.testdata.groups.COURSE_GROUP_MEMBERSHIPS['Extra Course']['Group Set 2']['extra-course-student-4'],
            ],
            None,
        ),
    ]

    # Strip down the queries (which may be partially resolved).
    def _clean_result(result):
        result = copy.deepcopy(result)

        for item in result:
            item.groupset.name = None
            item.group.name = None
            item.user.name = None
            item.user.email = None

        return result

    test.base_request_test(test.backend.courses_groups_memberships_list, test_cases,
            actual_clean_func = _clean_result,
            expected_clean_func = _clean_result)



import glob
import os
import typing

import edq.core.log
import edq.testing.cli
import edq.testing.unittest
import edq.testing.httpserver
import edq.util.pyimport

import lms.model.backend
import lms.model.base
import lms.backend.instance

THIS_DIR: str = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
TESTDATA_DIR: str = os.path.join(THIS_DIR, 'testdata')

BACKEND_TESTS_DIR: str = os.path.join(TESTDATA_DIR, 'backendtests')

CLI_TESTDATA_DIR: str = os.path.join(TESTDATA_DIR, 'cli')
CLI_TESTS_DIR: str = os.path.join(CLI_TESTDATA_DIR, 'tests')
CLI_DATA_DIR: str = os.path.join(CLI_TESTDATA_DIR, 'data')
CLI_GLOBAL_CONFG_PATH: str = os.path.join(CLI_DATA_DIR, 'testing-edq-lms.json')

TEST_FUNC_NAME_PREFIX: str = 'test_'
TEST_FILENAME_GLOB_PATTERN: str = '*_backendtest.py'

class BackendTest(edq.testing.httpserver.HTTPServerTest):
    """
    A special test suite that is common across all LMS backends.

    This is an HTTP test that will start a test server with exchanges specific to the target backend.

    A common directory (BACKEND_TESTS_DIR) will be searched for any file that starts with TEST_FILENAME_GLOB_PATTERN.
    Then, that file will be checked for any function that starts with TEST_FUNC_NAME_PREFIX and matches BackendTestFunction.
    """

    backend_type: typing.Union[str, None] = None
    """
    The backend type for this test.
    Must be set by the child class.
    """

    exchanges_dir: typing.Union[str, None] = None
    """
    The directory to load HTTP exchanges from.
    Must be set by the child class.
    """

    params_to_skip: typing.List[str] = []
    """ Parameters to skip while looking up exchanges. """

    headers_to_skip: typing.List[str] = []
    """ Headers to skip while looking up exchanges. """

    backend: typing.Union[lms.model.backend.APIBackend, None] = None
    """
    The backend for this test.
    Will be created during setup_server().
    """

    backend_args: typing.Dict[str, typing.Any] = {
        'testing': True,
    }
    """ Any additional arguments to send to get_backend(). """

    skip_base_request_test: bool = False
    """ Skip any base request tests. """

    allowed_backend: typing.Union[str, None] = None
    """ If set, skip any backend tests that do not match this filter. """

    override_server_url: typing.Union[str, None] = None
    """ If set, return this URL from get_server_url(). """

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)

        self._user_email: typing.Union[str, None] = None
        """
        The email of the current user for this backend.
        Setting the user allows child classes to fetch specific information (like authentication information).
        """

    @classmethod
    def setup_server(cls, server: edq.testing.httpserver.HTTPTestServer) -> None:
        if (cls.server_key == ''):
            raise ValueError("BackendTest subclass did not set server key properly.")

        if (cls.backend_type is None):
            raise ValueError("BackendTest subclass did not set backend type properly.")

        if (cls.exchanges_dir is None):
            raise ValueError("BackendTest subclass did not set exchanges dir properly.")

        edq.testing.httpserver.HTTPServerTest.setup_server(server)
        server.load_exchanges_dir(cls.exchanges_dir)

        # Update match options.
        for (key, values) in [('params_to_skip', cls.params_to_skip), ('headers_to_skip', cls.headers_to_skip)]:
            if (key not in server.match_options):
                server.match_options[key] = []

            server.match_options[key] += values

    @classmethod
    def post_start_server(cls, server: edq.testing.httpserver.HTTPTestServer) -> None:
        cls.backend = lms.backend.instance.get_backend(cls.get_server_url(), backend_type = cls.backend_type, **cls.backend_args)

    @classmethod
    def get_base_args(cls) -> typing.Dict[str, typing.Any]:
        """ Get a copy of the base arguments for a request (function). """

        return {}

    @classmethod
    def get_server_url(cls) -> str:
        if (cls.override_server_url is not None):
            return cls.override_server_url

        return str(super().get_server_url())

    def setUp(self) -> None:
        edq.core.log.init('ERROR')

        self.clear_user()

    def set_user(self, email: str) -> None:
        """
        Set the current user for this test.
        This can be especially useful for child classes that need to set information based on the user
        (like authentication headers).
        """

        self._user_email = email

    def clear_user(self) -> None:
        """
        Clear the current user for this test.
        This is automatically called before each test method.
        """

        self._user_email = None

    def base_request_test(self,
            request_function: typing.Callable,
            test_cases: typing.List[typing.Tuple[typing.Dict[str, typing.Any], typing.Any, typing.Union[str, None]]],
            stop_on_notimplemented: bool = True,
            actual_clean_func: typing.Union[typing.Callable, None] = None,
            expected_clean_func: typing.Union[typing.Callable, None] = None,
            assertion_func: typing.Union[typing.Callable, None] = None,
            ) -> None:
        """
        A common test for the base request functionality.
        Test cases are passed in as: `[(kwargs (and overrides), expected, error substring), ...]`.
        """

        if ((self.allowed_backend is not None) and (self.allowed_backend != self.backend_type)):
            self.skipTest(f"Backend {self.backend_type} has been filtered.")

        skip_reason = None

        for (i, test_case) in enumerate(test_cases):
            (extra_kwargs, expected, error_substring) = test_case

            with self.subTest(msg = f"Case {i}:"):
                kwargs = self.get_base_args()
                kwargs.update(extra_kwargs)

                try:
                    actual = request_function(**kwargs)
                except NotImplementedError as ex:
                    # We must handle this directly since we are in a subtest.
                    if (stop_on_notimplemented):
                        skip_reason = str(ex)
                        break

                    self.skipTest(f"Backend component not implemented: {str(ex)}.")
                except Exception as ex:
                    error_string = self.format_error_string(ex)
                    if (error_substring is None):
                        self.fail(f"Unexpected error: '{error_string}'.")

                    self.assertIn(error_substring, error_string, 'Error is not as expected.')

                    continue

                if (error_substring is not None):
                    self.fail(f"Did not get expected error: '{error_substring}'.")

                if (actual_clean_func is not None):
                    actual = actual_clean_func(actual)

                if (expected_clean_func is not None):
                    expected = expected_clean_func(expected)

                # If we expect a tuple, compare the tuple contents instead of the tuple itself.
                if (isinstance(expected, tuple)):
                    if (not isinstance(actual, tuple)):
                        raise ValueError(f"Expected results to be a tuple, found '{type(actual)}'.")

                    if (len(expected) != len(actual)):
                        raise ValueError(f"Result size mismatch. Expected: {len(expected)}, Actual: {len(actual)}.")
                else:
                    # Wrap the results in a tuple.
                    expected = (expected, )
                    actual = (actual, )

                for i in range(len(expected)):  # pylint: disable=consider-using-enumerate
                    expected_value = expected[i]
                    actual_value = actual[i]

                    if (assertion_func is not None):
                        assertion_func(expected_value, actual_value)
                    elif (isinstance(expected_value, lms.model.base.BaseType)):
                        self.assertJSONEqual(expected_value, actual_value)
                    elif (isinstance(expected_value, dict)):
                        self.assertJSONDictEqual(expected_value, actual_value)
                    elif (isinstance(expected_value, list)):
                        self.assertJSONListEqual(expected_value, actual_value)
                    else:
                        self.assertEqual(expected_value, actual_value)

        if (skip_reason is not None):
            self.skipTest(f"Backend component not implemented: {skip_reason}.")

    def modify_cli_test_info(self, test_info: edq.testing.cli.CLITestInfo) -> None:
        """ Adjust the CLI test info to include core info (like server information). """

        test_info.arguments += [
            '--config-global', CLI_GLOBAL_CONFG_PATH,
            '--server', self.get_server_url(),
            '--server-type', str(self.backend_type),
            '--config', 'testing=true',
        ]

        # Mark this CLI test for skipping based on the backend filter.
        if ((self.allowed_backend is not None) and (self.allowed_backend != self.backend_type)):
            test_info.skip_reasons.append(f"Backend {self.backend_type} has been filtered.")

    @classmethod
    def get_test_basename(cls, path: str) -> str:
        """ Get the test's name based off of its filename and location. """

        path = os.path.abspath(path)

        name = os.path.splitext(os.path.basename(path))[0]

        ancestors = os.path.dirname(path).replace(CLI_TESTS_DIR, '')
        prefix = ancestors.replace(os.sep, '_')

        if (prefix.startswith('_')):
            prefix = prefix.replace('_', '', 1)

        if (len(prefix) > 0):
            name =  f"{prefix}_{name}"

        return name

@typing.runtime_checkable
class BackendTestFunction(typing.Protocol):
    """
    A test function for backend tests.
    A copy of this function will be attached to a test class created for each backend.
    Therefore, `self` will be an instance of BackendTest.
    """

    def __call__(self, test: BackendTest) -> None:
        """
        A unit test for a BackendTest.
        """

def _wrap_test_function(test_function: BackendTestFunction) -> typing.Callable:
    """ Wrap the backend test function in some common code for backend tests. """

    def __method(self: BackendTest) -> None:
        try:
            test_function(self)
        except NotImplementedError as ex:
            # Skip tests for backend component that do not have implementations.
            self.skipTest(f"Backend component not implemented: {str(ex)}.")

    return __method

def add_test_path(target_class: type, path: str) -> None:
    """ Add tests from the given test files. """

    test_module = edq.util.pyimport.import_path(path)

    for attr_name in sorted(dir(test_module)):
        if (not attr_name.startswith(TEST_FUNC_NAME_PREFIX)):
            continue

        test_function = getattr(test_module, attr_name)
        setattr(target_class, attr_name, _wrap_test_function(test_function))

def discover_test_cases(target_class: type) -> None:
    """ Look in the text cases directory for any test cases and add them as test methods to the test class. """

    paths = list(sorted(glob.glob(os.path.join(BACKEND_TESTS_DIR, "**", TEST_FILENAME_GLOB_PATTERN), recursive = True)))
    for path in sorted(paths):
        add_test_path(target_class, path)

def attach_test_cases(target_class: type) -> None:
    """ Attach all the standard test cases to the given class. """

    # Attach backend tests.
    discover_test_cases(target_class)

    # Attach CLI tests.
    edq.testing.cli.discover_test_cases(target_class, CLI_TESTS_DIR, CLI_DATA_DIR, test_method_wrapper = _wrap_cli_test_method)

def _wrap_cli_test_method(test_method: typing.Callable, test_info_path: str) -> typing.Callable:
    """ Wrap the CLI tests to ignore NotImplemented errors. """

    def __method(self: edq.testing.unittest.BaseTest) -> None:
        try:
            test_method(self, reraise_exception_types = (NotImplementedError,))
        except NotImplementedError as ex:
            # Skip tests for backend component that do not have implementations.
            self.skipTest(f"Backend component not implemented: {str(ex)}.")

    return __method



import copy

import lms.backend.testing
import lms.model.courses
import lms.model.groups
import lms.model.groupsets
import lms.model.users
import lms.model.testdata.groups

DUMMY_ID: str = '123456789'

def test_courses_groupsets_memberships_resolve_and_set_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of resolving and setting group set memberships. """

    created_group_name = 'test_group_1'
    created_group = lms.model.groups.Group(id = DUMMY_ID, name = created_group_name)

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # No Op
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-2'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-3'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-4'),
                    ),
                ]
            },
            (
                [],
                [],
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 0,
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 0,
                },
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 0,
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 0,
                },
            ),
            None,
        ),

        # Clear Out
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                ]
            },
            (
                [],
                [
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(),
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(),
                ],
                {
                },
                {
                },
            ),
            None,
        ),

        # Move
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-1'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-2'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-3'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-4'),
                    ),
                ]
            },
            (
                [],
                [],
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 0,
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 1,
                },
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(): 1,
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 0,
                },
            ),
            None,
        ),

        # Create / Delete
        (
            {
                'course_query': lms.model.courses.CourseQuery(name = 'Extra Course'),
                'groupset_query': lms.model.groupsets.GroupSetQuery(name = 'Group Set 1'),
                'memberships': [
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = created_group_name),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-1'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-2'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-3'),
                    ),
                    lms.model.groups.GroupMembership(
                        group = lms.model.groups.GroupQuery(name = 'Group 1-2'),
                        user = lms.model.users.UserQuery(name = 'extra-course-student-4'),
                    ),
                ]
            },
            (
                [
                    created_group,
                ],
                [
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-1'].to_query(),
                ],
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 1,
                    created_group.to_query(): 1,
                },
                {
                    lms.model.testdata.groups.COURSE_GROUPS['Extra Course']['Group 1-2'].to_query(): 0,
                    created_group.to_query(): 0,
                },
            ),
            None,
        ),

    ]

    def _assertion_func(expected, actual):
        if (isinstance(expected, list)):
            test.assertJSONListEqual(expected, actual)
            return

        # Convert the non-JSON keys.

        clean_expected = {}
        for (key, value) in expected.items():
            clean_expected[str(key)] = value

        clean_actual = {}
        for (key, value) in actual.items():
            clean_actual[str(key)] = value

        test.assertDictEqual(clean_expected, clean_actual)

    # IDs from the backend may be inconsistent.
    def _clean_result(result):
        result = copy.deepcopy(result)

        # Clean the created groups.
        for group in result[0]:
            group.id = DUMMY_ID

        # Clean the added group counts.
        for group in list(result[2].keys()):
            if (group.name == created_group_name):
                group.id = DUMMY_ID

        return result

    test.base_request_test(test.backend.courses_groupsets_memberships_resolve_and_set, test_cases,
            assertion_func = _assertion_func,
            actual_clean_func = _clean_result)



"""
The `lms.cli` package contains general tools for interacting with Learning Management Systems (LMSs).
Each package can be invoked to list the tools (or subpackages) it contains.
Each tool includes a help prompt that accessed with the `-h`/`--help` flag.
"""

import sys

import edq.util.cli

def main() -> int:
    """ List this CLI dir. """

    return edq.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



"""
List the membership of a group.
"""

import argparse
import sys

import lms.backend.instance
import lms.cli.common
import lms.cli.parser
import lms.model.base

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    config = args._config

    backend = lms.backend.instance.get_backend(**config)

    course_query = lms.cli.common.check_required_course(backend, config)
    if (course_query is None):
        return 1

    groupset_query = lms.cli.common.check_required_groupset(backend, config)
    if (groupset_query is None):
        return 2

    group_query = lms.cli.common.check_required_group(backend, config)
    if (group_query is None):
        return 3

    memberships = backend.courses_groups_memberships_resolve_and_list(course_query, groupset_query, group_query)

    output = lms.model.base.base_list_to_output_format(memberships, args.output_format,
            skip_headers = args.skip_headers,
            pretty_headers = args.pretty_headers,
            include_extra_fields = args.include_extra_fields,
    )

    print(output)

    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    return run_cli(_get_parser().parse_args())

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    return lms.cli.parser.get_parser(__doc__.strip(),
            include_output_format = True,
            include_course = True,
            include_groupset = True,
            include_group = True,
    )

if (__name__ == '__main__'):
    sys.exit(main())



"""
Get the version of the LMS Toolkit package.
"""

import argparse
import sys

import lms
import lms.cli.parser

def run_cli(args: argparse.Namespace) -> int:
    """ Run the CLI. """

    print(f"v{lms.__version__}")
    return 0

def main() -> int:
    """ Get a parser, parse the args, and call run. """
    args, _ = _get_parser().parse_known_args()
    return run_cli(args)

def _get_parser() -> argparse.ArgumentParser:
    """ Get the parser. """

    return lms.cli.parser.get_parser(__doc__.strip(),
            include_server = False,
            include_auth = False,
    )

if (__name__ == '__main__'):
    sys.exit(main())



import typing

import edq.util.json

import lms.model.constants
import lms.util.string

TEXT_SEPARATOR: str = ': '
TEXT_EMPTY_VALUE: str = ''

T = typing.TypeVar('T', bound = 'BaseType')

class BaseType(edq.util.json.DictConverter):
    """
    The base class for all core LMS types.
    This class ensures that all children have the core functionality necessary for this package.

    The typical structure of types in this package is that types in the model package extend this class.
    Then, backends may declare their own types that extend the other classes from the model package.
    For example: lms.model.base.BaseType -> lms.model.assignments.Assignment -> lms.backend.canvas.model.assignments.Assignment

    General (but less efficient) implementations of core functions will be provided.
    """

    CORE_FIELDS: typing.List[str] = []
    """
    The common fields shared across backends for this type that are used for comparison and other operations.
    Child classes should set this to define how comparisons are made.
    """

    INT_COMPARISON_FIELDS: typing.Set[str] = {'id'}
    """
    Fields that should be compared like ints (even if they are strings).
    By default, this set will include 'id'.
    """

    def __init__(self,
            **kwargs: typing.Any) -> None:
        self.extra_fields: typing.Dict[str, typing.Any] = kwargs.copy()
        """ Additional fields not common to all backends or explicitly used by the creating child backend. """

    def __eq__(self, other: object) -> bool:
        if (not isinstance(other, BaseType)):
            return False

        # Check the specified fields only.
        for field_name in self.CORE_FIELDS:
            if (not hasattr(other, field_name)):
                return False

            value_self = getattr(self, field_name)
            value_other = getattr(other, field_name)

            if (field_name in self.INT_COMPARISON_FIELDS):
                comparison = lms.util.string.compare_maybe_ints(value_self, value_other)
                if (comparison != 0):
                    return False
            elif (value_self != value_other):
                return False

        return True

    def __hash__(self) -> int:
        values = tuple(getattr(self, field_name) for field_name in self.CORE_FIELDS)
        return hash(values)

    def __lt__(self, other: 'BaseType') -> bool:  # type: ignore[override]
        if (not isinstance(other, BaseType)):
            return False

        # Check the specified fields only.
        for field_name in self.CORE_FIELDS:
            if (not hasattr(other, field_name)):
                return False

            value_self = getattr(self, field_name)
            value_other = getattr(other, field_name)

            if (field_name in self.INT_COMPARISON_FIELDS):
                comparison = lms.util.string.compare_maybe_ints(value_self, value_other)
                if (comparison == 0):
                    continue

                return (comparison < 0)

            if (value_self == value_other):
                continue

            return bool(value_self < value_other)

        return False

    def as_text_rows(self,
            skip_headers: bool = False,
            pretty_headers: bool = False,
            **kwargs: typing.Any) -> typing.List[str]:
        """
        Create a representation of this object in the "text" style of this project meant for display.
        A list of rows will be returned.
        """

        rows = []

        for (field_name, row) in self._get_fields(**kwargs).items():
            if (not skip_headers):
                header = field_name
                if (pretty_headers):
                    header = header.replace('_', ' ').title()

                row = f"{header}{TEXT_SEPARATOR}{row}"

            rows.append(row)

        return rows

    def get_headers(self,
            pretty_headers: bool = False,
            **kwargs: typing.Any) -> typing.List[str]:
        """
        Get a list of headers to label the values represented by this object meant for display.
        This method is a companion to as_table_rows(),
        given the same options these two methods will produce rows with the same length and ordering.
        """

        headers = []

        for field_name in self._get_fields(**kwargs):
            header = field_name
            if (pretty_headers):
                header = header.replace('_', ' ').title()

            headers.append(header)

        return headers

    def as_table_rows(self,
            **kwargs: typing.Any) -> typing.List[typing.List[str]]:
        """
        Get a list of the values by this object meant for display.
        This method is a companion to get_headers(),
        given the same options these two methods will produce rows with the same length and ordering.

        Note that the default implementation for this method always return a single row,
        but children may override and return multiple rows per object.
        """

        return [list(self._get_fields(**kwargs).values())]

    def as_json_dict(self,
            **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        """
        Get a dict representation of this object meant for display as JSON.
        (Note that we are not returning JSON, just a dict that is ready to be converted to JSON.)
        Calling this method differs from passing this object to json.dumps() (or any sibling),
        because this method may not include all fields, may flatten or alter fields, and will order fields differently.
        """

        return {field_name: self._get_field_value(field_name) for field_name in self._get_fields(**kwargs)}

    def _get_fields(self,
            include_extra_fields: bool = False,
            **kwargs: typing.Any) -> typing.Dict[str, str]:
        """
        Get a dictionary representing the "target" fields of this object meant for display.
        Keys (field names) will not be modified, but values will be sent to self._value_to_text().
        Keys are placed in the dictionary in a consistent ordering.
        """

        field_names = self.CORE_FIELDS.copy()

        # Append any extra fields after the core fields.
        if (include_extra_fields):
            # First, include any fields that are not in self.extra_fields.
            for extra_name in (list(vars(self).keys()) + list(self.extra_fields.keys())):
                if (extra_name == 'extra_fields'):
                    continue

                if (extra_name not in field_names):
                    field_names.append(extra_name)

        fields = {}
        for field_name in field_names:
            fields[field_name] = self._value_to_text(self._get_field_value(field_name), **kwargs)

        return fields

    def _get_field_value(self, name: str, default: typing.Any = None) -> typing.Any:
        """
        Get the value for a field.
        This is similar to `getattr(self, name, default)`,
        but this will also check `extra_fields` if the field is not found at the top level.
        """

        if (hasattr(self, name)):
            return getattr(self, name)

        if (name in self.extra_fields):
            return self.extra_fields[name]

        return default

    def _value_to_text(self,
            value: typing.Any,
            indent: typing.Union[int, None] = None,
            **kwargs: typing.Any) -> str:
        """
        Convert some arbitrary value (usually found within a BaseType) to a string.
        None values will be returned as `TEXT_EMPTY_VALUE`.
        """

        if (value is None):
            return TEXT_EMPTY_VALUE

        if (hasattr(value, '_to_text')):
            return str(value._to_text())

        if (isinstance(value, (edq.util.json.DictConverter, dict, list, tuple))):
            return str(edq.util.json.dumps(value, indent = indent))

        return str(value)

    @classmethod
    def from_json_dict(cls: typing.Type[T],
            data: typing.Dict[str, typing.Any],
            **kwargs: typing.Any) -> T:
        """
        Create an object from a dict that can be used for JSON.
        This is the inverse of as_json_dict().
        """

        return typing.cast(T, cls.from_dict(data))

def base_list_to_output_format(values: typing.Sequence[BaseType], output_format: str,
        sort: bool = True,
        skip_headers: bool = False,
        pretty_headers: bool = False,
        include_extra_fields: bool = False,
        **kwargs: typing.Any) -> str:
    """
    Convert a list of base types to a string representation.
    The returned string will not include a trailing newline.

    The given list may be modified by this call.
    """

    values = list(values)

    if (sort):
        values.sort()

    output = ''

    if (output_format == lms.model.constants.OUTPUT_FORMAT_JSON):
        output = base_list_to_json(values,
                include_extra_fields = include_extra_fields,
                **kwargs)
    elif (output_format == lms.model.constants.OUTPUT_FORMAT_TABLE):
        output = base_list_to_table(values,
                skip_headers = skip_headers, pretty_headers = pretty_headers,
                include_extra_fields = include_extra_fields,
                **kwargs)
    elif (output_format == lms.model.constants.OUTPUT_FORMAT_TEXT):
        output = base_list_to_text(values,
                skip_headers = skip_headers, pretty_headers = pretty_headers,
                include_extra_fields = include_extra_fields,
                **kwargs)
    else:
        raise ValueError(f"Unknown output format: '{output_format}'.")

    return output

def base_list_to_json(values: typing.Sequence[BaseType],
        indent: int = 4,
        extract_single_list: bool = False,
        **kwargs: typing.Any) -> str:
    """ Convert a list of base types to a JSON string representation. """

    output_values = [value.as_json_dict(**kwargs) for value in values]
    if (extract_single_list and (len(output_values) == 1)):
        output_values = output_values[0]  # type: ignore[assignment]

    return str(edq.util.json.dumps(output_values, indent = indent, sort_keys = False))

def base_list_to_table(values: typing.Sequence[BaseType],
        skip_headers: bool = False,
        delim: str = "\t",
        **kwargs: typing.Any) -> str:
    """ Convert a list of base types to a table string representation. """

    rows = []

    if ((len(values) > 0) and (not skip_headers)):
        rows.append(values[0].get_headers(**kwargs))

    for value in values:
        rows += value.as_table_rows(**kwargs)

    return "\n".join([delim.join(row) for row in rows])

def base_list_to_text(values: typing.Sequence[BaseType],
        **kwargs: typing.Any) -> str:
    """ Convert a list of base types to a text string representation. """

    output = []

    for value in values:
        rows = value.as_text_rows(**kwargs)
        output.append("\n".join(rows))

    return "\n\n".join(output)



import lms.backend.testing
import lms.model.testdata.scores

def test_courses_assignments_scores_upload_base(test: lms.backend.testing.BackendTest):
    """ Test the base functionality of uploading assignments scores. """

    # [(kwargs (and overrides), expected, error substring), ...]
    test_cases = [
        # Base
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
                'scores': {
                    '6': lms.model.scores.ScoreFragment(score = 1.0),
                },
            },
            1,
            None,
        ),

        # Empty
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
                'scores': {
                    '6': lms.model.scores.ScoreFragment(),
                },
            },
            1,
            None,
        ),

        # Comment
        (
            {
                'course_id': '110000000',
                'assignment_id': '110000100',
                'scores': {
                    '6': lms.model.scores.ScoreFragment(score = 1.0, comment = 'foo'),
                },
            },
            1,
            None,
        ),
    ]

    test.base_request_test(test.backend.courses_assignments_scores_upload, test_cases)



"""
The `lms.model` package provides the base types used throughout this project.
Most of these types are written to be general,
and may be extended by specific backends to include information specific to their LMS.
"""



import os
import typing

import edq.testing.cli
import edq.util.net

import lms.backend.testing
import lms.model.constants

THIS_DIR: str = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
ROOT_DIR: str = os.path.join(THIS_DIR, '..', '..', '..')

BLACKBOARD_TEST_EXCHANGES_DIR: str = os.path.join(ROOT_DIR, 'testdata', 'lms-blackboard-testdata', 'testdata', 'http')

DEFAULT_USER: str = 'course-owner'

class BlackboardBackendTest(lms.backend.testing.BackendTest):
    """ A backend test for Blackboard. """

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)

        # Skip exchange verification tests, since Blackboard users are not logged in ahead of time.
        # This makes the exchanges dependent on the login.
        self.skip_test_exchanges_base = True

    @classmethod
    def child_class_setup(cls) -> None:
        cls.server_key = lms.model.constants.BACKEND_TYPE_BLACKBOARD

        cls.backend_type = lms.model.constants.BACKEND_TYPE_BLACKBOARD

        cls.exchanges_dir = BLACKBOARD_TEST_EXCHANGES_DIR

        cls.backend_args.update({
            'auth_user': DEFAULT_USER,
            'auth_password': DEFAULT_USER,
        })

        cls.params_to_skip += [
            'blackboard.platform.security.NonceUtil.nonce.ajax',
        ]

        cls.headers_to_skip += [
            'edq-lms-blackboard-user',
            'x-blackboard-xsrf',
        ]

        edq.util.net._disable_https_verification()

    def modify_cli_test_info(self, test_info: edq.testing.cli.CLITestInfo) -> None:
        super().modify_cli_test_info(test_info)

        test_info.arguments += [
            '--auth-user', self.backend._username,
            '--auth-password', self.backend._password,
        ]

    def set_user(self, email: str) -> None:
        super().set_user(email)

        username = email.split('@')[0]

        # Remember that test passwords are the same as usernames.
        self.backend._username = username
        self.backend._password = username

# Attatch tests to this class.
lms.backend.testing.attach_test_cases(BlackboardBackendTest)
